<?php
namespace WCStepFilter;

/**
 * Core Class
 *
 * @class Core
 * @version 5.2.1
 */
class Core
{
    // <editor-fold desc="Properties">
    /**
     * Self instance variable
     * @var Core The single instance of the class
     */
    protected static $instance = null;

    /**
     * Is WooCommerce active
     * @var bool
     */
    public static $wcIsActive = false;

    /**
     * Current working filter ID
     * @var int
     */
    public $currentId = null;

    /**
     * Current working question ID
     * @var int
     */
    public $currentQuestionId = null;

    /**
     * Current working filter results PDF total pages
     * @var int
     */
    public $currentResultsPDFTotalPages = null;

    /**
     * Plugin global settings model
     *
     * @var array
     */
    public static $settingsModel;

    /**
     * Array of the showed templates
     * @var array
     */
    public static $templatesShowed = [];

    /**
     * Array of the showing template args
     * @var array
     */
    public static $currentTemplateArgs = [];

    /**
     * Storage class instance
     *
     * @var object
     */
    public $storage;

    /**
     * Filter class instance
     *
     * @var object
     */
    public $filter;

    /**
     * Question class instance
     *
     * @var object
     */
    public $question;

    /**
     * Inquiry class instance
     *
     * @var object
     */
    public $inquiry;

    /**
     * Product attribute value settings model
     *
     * @var object
     */
    public static $attributeValueSettingsModel;

    /**
     * Product category settings model
     *
     * @var object
     */
    public static $categorySettingsModel;
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Constructor */
    public function __construct()
    {
        // include base slave classes
        $requiredClasses = [
            'L10N',
            'Utils',
            'Storage',
            'Filter',
            'Question',
            'Inquiry'
        ];

        foreach ($requiredClasses as $requiredClass) {
            if (!class_exists('\\WCStepFilter\\' . $requiredClass)
                && file_exists(__DIR__ . DIRECTORY_SEPARATOR . $requiredClass . '.php')
            ) {
                include_once(__DIR__ . DIRECTORY_SEPARATOR . $requiredClass . '.php');
            }
        }

        // actions
        add_action('init', [$this, 'onInit'], 5);
        add_action('plugins_loaded', [$this, 'pluginsLoadedHook']);
        add_action('woocommerce_init', [$this, 'onWoocommerceInit'], 1);
        add_action('wp_loaded', [$this, 'requests']);
        add_action('plugins_loaded', [$this, 'loadTextDomain']);
        add_action('woocommerce_before_shop_loop', [$this, 'outputShopControls']);
        add_action('woocommerce_no_products_found', [$this, 'outputShopControls']);
        add_action('wcsf_before_output', [$this, 'beforeOutputAction']);
        add_action('wcsf_after_output', [$this, 'afterOutputAction']);

        if (isset($_GET['wcsf-results'])) {
            add_action('woocommerce_product_query', [$this, 'modifyProductsQuery'], 15);
            add_filter('woocommerce_layered_nav_link', [$this, 'layeredNavLinkFilter']);
            add_filter('woocommerce_widget_get_current_page_url', [$this, 'layeredNavLinkFilter']);

            // filtering WOOF plugin args
            add_filter('woof_products_query', [$this, 'modifyWoofQuery'], 15);
        }

        // settings page
        add_filter('woocommerce_settings_tabs_array', [$this, 'addSettingsTab'], 50);
        add_action('woocommerce_settings_tabs_step_filter_settings_tab', [$this, 'initSettingsTab']);
        add_action('woocommerce_update_options_step_filter_settings_tab', [$this, 'updateSettingsTab']);

        // scripts and styles
        add_action('wp_enqueue_scripts', [$this, 'enqueueFront']);
        add_action('admin_enqueue_scripts', [$this, 'enqueueAdmin']);

        // ajax
        add_action('wp_ajax_woocommerce_json_search_categories', [$this, 'jsonSearchCategories']);
        add_action('wp_ajax_woocommerce_json_search_tags', [$this, 'jsonSearchTags']);
        add_action('wp_ajax_wcsfGetAttributeValuesSelectOptions', [$this, 'getAttributeValuesSelectOptionsAjax']);

        // init basic classes
        $this->storage = new Storage();
        $this->inquiry = new Inquiry();
        $this->filter = new Filter();
        $this->question = new Question();
    }

    /**
     * Get single class instance
     *
     * @static
     * @return Core
     */
    public static function instance()
    {
        if (self::$instance === null) {
            self::$instance = new self();
        }

        return self::$instance;
    }

    /** On init action */
    public function onInit()
    {
        self::$settingsModel = apply_filters(
            'wcsf_global_settings_model',
            [
                'styles_including_type' => [
                    'name' => L10N::r('Styles including type'),
                    'default' => 'full',
                    'type' => 'select',
                    'desc' => L10N::r('For more info see the')
                        . ' <a href="' . WC_STEP_FILTER_PLUGIN_URL
                        . 'documentation/index.html#settings" target="_blank">'
                        . L10N::r('documentation') . '</a>',
                    'id' => 'woocommerce_step_filter_styles_including_type',
                    'key' => 'woocommerce_step_filter_styles_including_type',
                    'options' => [
                        'full' => L10N::r('Full styles file'),
                        'basic' => L10N::r('Basic styles file'),
                        'none' => L10N::r('None')
                    ]
                ],
                'scripts_including_type' => [
                    'name' => L10N::r('Scripts including type'),
                    'default' => 'single',
                    'type' => 'select',
                    'desc' => 'Include scripts separately or in a single file',
                    'id' => 'woocommerce_step_filter_scripts_including_type',
                    'key' => 'woocommerce_step_filter_scripts_including_type',
                    'options' => [
                        'single' => L10N::r('Single file'),
                        'multiple' => L10N::r('Multiple files')
                    ]
                ],
                'included_scripts' => [
                    'name' => L10N::r('Included scripts'),
                    'default' => [
                        'app',
                        'hooks',
                        'wNumb',
                        'nouislider',
                        'nouislider-launcher'
                    ],
                    'type' => 'multiselect',
                    'desc' => 'Select which files will be included with the multiple including type',
                    'id' => 'woocommerce_step_filter_included_scripts',
                    'key' => 'woocommerce_step_filter_included_scripts',
                    'options' => [
                        'app' => 'app',
                        'hooks' => 'hooks',
                        'wNumb' => 'wNumb',
                        'nouislider' => 'noUiSlider',
                        'nouislider-launcher' => 'noUiSlider-launcher'
                    ]
                ],
                'store_session_in_db' => [
                    'name' => L10N::r('Store session in the DB'),
                    'default' => 'no',
                    'type' => 'checkbox',
                    'desc' => L10N::r('Tick in a case of problems of the filter\'s state storing'),
                    'id' => 'woocommerce_step_filter_store_session_in_db',
                    'key' => 'woocommerce_step_filter_store_session_in_db'
                ]
            ]
        );

        self::$attributeValueSettingsModel = apply_filters(
            'wcsf_attribute_value_settings_model',
            [
                'pre_checked' => [
                    'label' => L10N::r('Step Filter Pre-checked'),
                    'key' => '_wcsf_pre_checked',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'thumbnail' => [
                    'label' => L10N::r('Step Filter Thumbnail'),
                    'key' => '_wcsf_thumbnail',
                    'type' => 'thumbnail',
                    'default' => ''
                ]
            ]
        );

        self::$categorySettingsModel = apply_filters(
            'wcsf_category_settings_model',
            [
                'pre_checked' => [
                    'label' => L10N::r('Step Filter Pre-checked'),
                    'key' => '_wcsf_pre_checked',
                    'type' => 'checkbox',
                    'default' => false
                ]
            ]
        );
    }

    /** Handles then plugins are loaded */
    public function pluginsLoadedHook()
    {
        self::$wcIsActive = class_exists('\WooCommerce');

        if (!self::$wcIsActive) {
            add_action('admin_notices', function () {
                echo '<div class="notice notice-warning is-dismissible"><p>'
                    . L10N::r('WooCommerce is required for WC Step Filter')
                    . '</p></div>';
            });
        }
    }

    /** Runs after woocommerce is loaded */
    public function onWoocommerceInit()
    {
        add_action('wcsf_shortcode', [$this, 'shortCodeAction']);
        add_action('product_cat_edit_form_fields', [$this, 'categoryFieldsAdd'], 10, 2);
        add_action('product_tag_edit_form_fields', [$this, 'categoryFieldsAdd'], 10, 2);
        add_action('edited_product_cat', [$this, 'categoryFieldsSave'], 10, 2);
        add_action('edited_product_tag', [$this, 'categoryFieldsSave'], 10, 2);

        $attributes = \wc_get_attribute_taxonomies();

        foreach ($attributes as $attribute) {
            add_action(
                'pa_' . $attribute->attribute_name . '_edit_form_fields',
                [$this, 'attributeValueFieldsAdd'],
                10,
                2
            );

            add_action('edited_pa_' . $attribute->attribute_name, [$this, 'attributeValueFieldsSave'], 10, 2);
        }

        // start session if it doesn't exist
        if (!session_id() && !(is_admin() && !is_ajax())) {
            @session_start();
        }

        // start WC session variable if needed
        if (self::getSetting('store_session_in_db') && function_exists('WC')) {
            if (method_exists(WC(), 'initialize_session')) {
                WC()->initialize_session();
            }

            if (method_exists(WC()->session, 'set_customer_session_cookie')) {
                WC()->session->set_customer_session_cookie(true);
            }
        }
    }

    /**
     * Short-code call action
     *
     * @param array $attributes
     */
    public function shortCodeAction($attributes)
    {
        if (!isset($attributes['id']) || empty($attributes['id'])) {
            return;
        }

        $id = (int) $attributes['id'];
        $resultsUrl = $this->filter->getSetting($id, 'results_url');

        if ($resultsUrl == '#') {
            if (!isset($_GET['wcsf-results']) || empty($_GET['wcsf-results'])) {
                // reset filter's state if it's results page is any and there is no query parameters
                $this->filter->resetState($id);
            } elseif (isset($_GET['wcsf-value'], $_GET['wcsf-results']) && !empty($_GET['wcsf-value'])
                && $_GET['wcsf-results'] == $id
            ) {
                try {
                    // set filter's value from get parameter
                    $this->filter->setValue($id, $_GET['wcsf-value']);
                } catch (\Exception $exception) {
                    $this->filter->addNotice(
                        $exception->getCode() ? $exception->getCode() : $this->filter->getActiveStepId($id),
                        ['message' => $exception->getMessage()]
                    );
                }
            }
        }
    }

    /**
     * Handles on output call
     *
     * @param array $args
     */
    public function beforeOutputAction($args)
    {
        if (isset($args['id']) && $args['id']) {
            $this->currentId = (int) $args['id'];
        }
    }

    /** Handles after output call */
    public function afterOutputAction()
    {
        $this->currentId = null;
    }

    /** Load text domain */
    public function loadTextDomain()
    {
        load_plugin_textdomain(
            'woocommerce-step-filter',
            false,
            basename(WC_STEP_FILTER_PLUGIN_PATH) . '/languages/'
        );
    }
    // </editor-fold>

    // <editor-fold desc="HTML">
    /** Output shop controls html */
    public function outputShopControls()
    {
        if (isset($_GET['wcsf-results']) && !empty($_GET['wcsf-results'])) {
            self::getTemplatePart(
                'shop-controls/index',
                ['id' => (int) $_GET['wcsf-results']],
                ['once' => true]
            );
        }
    }

    /**
     * Include php-template by the name.
     * First looking in the "theme folder/woocommerce-step-filter (WC_STEP_FILTER_THEME_TEMPLATES_DIR)"
     * Second looking in the "plugin folder/views"
     * Making extraction of the arguments as variables
     *
     * @param string $name
     * @param array  $arguments
     * @param array  $templateSettings
     *
     * @return string
     */
    public static function getTemplatePart($name = '', $arguments = [], $templateSettings = [])
    {
        $defaultSettings = [
            'echo' => true,
            'once' => false,
            'compress' => !WC_STEP_FILTER_DEBUG
        ];

        $templateSettings = array_merge($defaultSettings, $templateSettings);

        // show template only once
        if ($templateSettings['once'] && in_array($name, self::$templatesShowed)) {
            return null;
        }

        // save template as showed
        self::$templatesShowed[] = $name;

        if (is_array($arguments)) {
            extract($arguments, EXTR_PREFIX_SAME, 'data');

            self::$currentTemplateArgs = $arguments;
        }

        $path = get_stylesheet_directory() . DIRECTORY_SEPARATOR
            . WC_STEP_FILTER_THEME_TEMPLATES_DIR . DIRECTORY_SEPARATOR . $name . '.php';

        if (!file_exists($path)) {
            $path = WC_STEP_FILTER_PLUGIN_PATH . DIRECTORY_SEPARATOR
                . 'views' . DIRECTORY_SEPARATOR . $name . '.php';
        }

        if (!file_exists($path)) {
            return '';
        }

        ob_start();

        include($path);

        $html = ob_get_clean();

        if ($templateSettings['compress']) {
            $replace = [
                '/\>[^\S ]+/s' => '>',      // strip whitespaces after tags, except space
                '/[^\S ]+\</s' => '<',      // strip whitespaces before tags, except space
                '/(\s)+/s' => '\\1',        // shorten multiple whitespace sequences
                '/<!--spacer-->/' => ' ',   // replace spacer tag
                '/<!--(.|\s)*?-->/' => ''   // remove HTML comments
            ];

            $html = preg_replace(array_keys($replace), array_values($replace), $html);
        }

        if ($templateSettings['echo']) {
            echo $html;

            return null;
        }

        return $html;
    }

    /**
     * Get requested HTML part arguments
     *
     * @param array $defaults
     * @param array $settings
     *
     * @return array
     */
    public static function getTemplatePartArgs($defaults = [], $settings = [])
    {
        $defaultsSettings = ['recursive' => false];
        $settings = array_replace($defaultsSettings, $settings);
        $arguments = self::$currentTemplateArgs;

        if (!empty($arguments)) {
            foreach ($defaults as $defaultKey => $_) {
                // find arguments from shortcode attributes
                if (strtolower($defaultKey) == $defaultKey || !isset($arguments[strtolower($defaultKey)])) {
                    continue;
                }

                $arguments[$defaultKey] = $arguments[strtolower($defaultKey)];

                unset($arguments[strtolower($defaultKey)]);
            }

            if ($settings['recursive']) {
                $arguments = array_replace_recursive($defaults, $arguments);
            } else {
                $arguments = array_replace($defaults, $arguments);
            }

            return $arguments;
        }

        return $defaults;
    }

    /**
     * Get available nav list templates from plugin and theme directory
     *
     * @return array $output
     */
    public static function getNavViewsList()
    {
        $output = [];
        $pluginFiles = WC_STEP_FILTER_PLUGIN_PATH . DIRECTORY_SEPARATOR . 'views' . DIRECTORY_SEPARATOR . 'nav'
            . DIRECTORY_SEPARATOR . 'list';

        $themeFiles = get_stylesheet_directory() . DIRECTORY_SEPARATOR . WC_STEP_FILTER_THEME_TEMPLATES_DIR
            . DIRECTORY_SEPARATOR . 'nav' . DIRECTORY_SEPARATOR . 'list';

        foreach ([$pluginFiles, $themeFiles] as $sourceFiles) {
            if (!file_exists($sourceFiles)) {
                continue;
            }

            foreach (scandir($sourceFiles) as $file) {
                if (is_dir($sourceFiles . DIRECTORY_SEPARATOR . $file)) {
                    continue;
                }

                $name = str_replace('.php', '', $file);
                $output[$name] = $name;
            }
        }

        return apply_filters('wcsf_nav_list_views', $output);
    }
    // </editor-fold>

    // <editor-fold desc="Query">
    /** Add request actions */
    public function requests()
    {
        if (is_admin()) {
            return;
        }

        // get queries
        // reset the filter
        if (!isset($_POST['results']) && isset($_GET['wcsf-reset-state'])) {
            $this->filter->resetState((int) $_GET['wcsf-reset-state']);
        }

        // reset one value
        if (!isset($_POST['results']) && isset($_GET['wcsf-reset-value'], $_GET['wcsf-results'])) {
            $this->filter->resetValue((int) $_GET['wcsf-results'], (int) $_GET['wcsf-reset-value']);
        }

        // post queries
        if (!isset($_POST['wcsf'])) {
            return;
        }

        $id = isset($_POST['id']) ? (int) $_POST['id'] : false;

        // skip step
        if (isset($_POST['skip'])) {
            $this->filter->setActiveStep($id, esc_sql($_POST['skip']));

            return;
        }

        // back step
        if (isset($_POST['back'])) {
            $this->filter->setActiveStep($id, esc_sql($_POST['back']));

            return;
        }

        // save value
        if (isset($_POST['value']) && !empty($_POST['value'])) {
            try {
                $this->filter->setValue($id, (array) $_POST['value']);
            } catch (\Exception $exception) {
                $this->filter->addNotice(
                    $exception->getCode() ? $exception->getCode() : $this->filter->getActiveStepId($id),
                    ['message' => $exception->getMessage()]
                );
            }
        }

        // change step
        if (isset($_POST['stepId'])) {
            $this->filter->setActiveStep($id, esc_sql($_POST['stepId']));
        }

        // reset filter
        if (isset($_POST['reset'])) {
            $this->filter->resetState($id);
        }

        // show results page
        if (isset($_POST['results'])) {
            $redirectLink = $this->filter->getResultsUrl($id);
            $redirectLink = apply_filters('wcsf_results_redirect_link', $redirectLink, $id);
            $redirectCode = apply_filters('wcsf_results_redirect_code', 302, $id, $redirectLink);

            do_action('wcsf_before_results_redirect', $id, $redirectLink);

            wp_redirect($redirectLink, $redirectCode);

            // js version of redirect
            exit("<script>document.location = '$redirectLink';</script>");
        }
    }

    /**
     * Execute question formula by values
     *
     * @param object $question - question object
     * @param array $values - question values
     *
     * @return float
     *
     * @throws \Exception
     */
    public static function executeFormula($question, $values)
    {
        if (!class_exists('\Matex\Evaluator')) {
            require_once __DIR__ . '/../vendor/matex/Evaluator.php';
            require_once __DIR__ . '/../vendor/matex/Math.php';
        }

        $evaluator = new \Matex\Evaluator();
        $evaluator->variables = [];
        $evaluator->functions = \Matex\Math::functions;

        foreach ($question->formulaVariables as $variable) {
            $evaluator->variables[$variable['name']] = (float) $variable['value'];
        }

        foreach ($values as $fieldName => $fieldValue) {
            foreach ($question->formulaValueAliases as $alias) {
                if ($alias['name'] == $fieldName && $alias['value'] == $fieldValue) {
                    $fieldValue = $alias['alias'];

                    continue;
                }
            }

            foreach ($question->formulaNameAliases as $alias) {
                if ($alias['name'] == $fieldName) {
                    $fieldName = $alias['alias'];

                    continue;
                }
            }

            $evaluator->variables[$fieldName] = (float) $fieldValue;
        }

        return $evaluator->execute($question->formula);
    }

    /**
     * Get query args by the filter id and value
     *
     * @param integer $id
     * @param array $filterValue
     * @param array $args
     *
     * @return array
     */
    public static function getQueryArgsByFilter($id, $filterValue, $args = [])
    {
        $taxQuery = [];
        $postIn = [];
        $postNotIn = [];
        $metaQuery = [];
        $extraQueryArgs = Filter::getSetting($id, 'extra_query_args');
        $multiChoiceOperators = apply_filters(
            'wsf_multi_choice_operators',
            [
                -1 => 'NOT IN',
                1 => 'IN',
                2 => 'AND'
            ]
        );

        if (!empty($extraQueryArgs)) {
            foreach ($extraQueryArgs as $extraArg) {
                if (($extraArg['source'] == 'product' && empty($extraArg['product']))
                    || ($extraArg['source'] == 'category' && empty($extraArg['category']))
                    || ($extraArg['source'] == 'attribute'
                        && (empty($extraArg['attribute_name']) || empty($extraArg['attribute_value'])))
                ) {
                    continue;
                }

                switch ($extraArg['source']) {
                    case 'product':
                        if ($extraArg['condition'] == '=') {
                            $postIn[] = (int) $extraArg['product'];
                        } else {
                            $postNotIn[] = (int) $extraArg['product'];
                        }

                        break;
                    case 'category':
                        $taxQuery[] = [
                            'taxonomy' => 'product_cat',
                            'field' => 'id',
                            'terms' => (int) $extraArg['category'],
                            'operator' => $extraArg['condition'] == '=' ? 'IN' : 'NOT IN'
                        ];

                        break;
                    case 'attribute':
                        $taxQuery[] = [
                            'taxonomy' => 'pa_' . esc_sql($extraArg['attribute_name']),
                            'field' => 'id',
                            'terms' => (int) $extraArg['attribute_value'],
                            'operator' => $extraArg['condition'] == '=' ? 'IN' : 'NOT IN'
                        ];
                }
            }
        }

        // blend in passed category or tag id
        if (isset($args['categoryId']) && $args['categoryId']) {
            $taxQuery[] = [
                'taxonomy' => 'product_cat',
                'field' => 'id',
                'terms' => $args['categoryId']
            ];
        } elseif (isset($args['tagId']) && $args['tagId']) {
            $taxQuery[] = [
                'taxonomy' => 'product_tag',
                'field' => 'id',
                'terms' => $args['tagId']
            ];
        }

        // loop for the each filter value
        foreach ((array) $filterValue as $questionId => $questionValue) {
            $question = Question::get($questionId);

            if (!is_object($question) || !property_exists($question, 'viewType') || $question->ignoreFiltering) {
                continue;
            }

            $viewType = $question->viewType;
            $filterType = $question->filterType;
            $relation = isset($question->relation) ? $question->relation : 'or';
            $operator = $relation == 'or' ? 'IN' : 'AND';
            $formulaValue = 0;

            if ($viewType == 'formula') {
                try {
                    $formulaValue = self::executeFormula($question, $questionValue);
                } catch (\Exception $exception) {
                    echo '<p class="wcsf-question-formula-error" role="alert"><strong>'
                        . L10N::r('Formula execute error:')
                        . " <em>{$exception->getMessage()}</em></strong></p>";

                    continue;
                }
            }

            if (in_array($filterType, ['tag', 'category', 'attribute'])) {
                // attribute, category or tag
                if ($filterType == 'attribute') {
                    // filtering by attribute
                    if (!isset($question->attribute)) {
                        continue;
                    }

                    $taxonomy = \wc_attribute_taxonomy_name_by_id((int) $question->attribute);
                } else {
                    // filtering by tag or category
                    $taxonomy = $filterType == 'tag' ? 'product_tag' : 'product_cat';
                }

                if (strpos($viewType, 'number') !== false) {
                    // numeric field type
                    $filterBy = [];
                    $terms = get_terms(['taxonomy' => $taxonomy]);
                    $valueSource = Question::getSetting($questionId, 'attribute_value_source');

                    foreach ($terms as $term) {
                        switch ($valueSource) {
                            default:
                            case 'name':
                                $termValue = (float) $term->name;
                                break;

                            case 'slug':
                                $termValue = (float) $term->slug;
                        }

                        $issetFrom = isset($questionValue['from']) && !empty($questionValue['from']);
                        $from = $issetFrom && $questionValue['from'] <= $termValue;
                        $issetTo = isset($questionValue['to']) && !empty($questionValue['to']);
                        $to = $issetTo && $questionValue['to'] >= $termValue;

                        // check by numeric with the first, the second and both values
                        if (($viewType == 'number' && !empty($questionValue))
                            || ($viewType == 'number-from' && $from)
                            || ($viewType == 'number-to' && $to)
                            || ($viewType == 'number-between'
                                && ($from && $to || !$issetTo && $from || !$issetFrom && $to))
                        ) {
                            $filterBy[] = $term->term_id;
                        }
                    }

                    $taxQuery[] = [
                        'taxonomy' => $taxonomy,
                        'field' => 'id',
                        'terms' => $filterBy,
                        'operator' => 'IN'
                    ];
                } else {
                    if (filter_var(get_option('woocommerce_hide_out_of_stock_items'), FILTER_VALIDATE_BOOLEAN)
                        && $filterType == 'attribute'
                        && $viewType != 'formula'
                    ) {
                        // skip products with out-of-stock variations by requested attributes
                        global $wpdb;

                        $visibilityTerms = wc_get_product_visibility_term_ids();
                        $include = array_filter($questionValue);

                        if (strpos($viewType, 'multi-choice') !== false) {
                            $include = [];

                            foreach (array_filter($questionValue) as $termId => $choiceValue) {
                                if ($choiceValue > 0) {
                                    $include[] = $termId;
                                }
                            }
                        }

                        // using of an SQL query avoids cross wp_queries problems
                        $queryString =
                            "SELECT $wpdb->posts.post_parent, $wpdb->term_relationships.term_taxonomy_id "
                            . "FROM $wpdb->posts "
                            . "LEFT JOIN $wpdb->term_relationships "
                            . "ON ($wpdb->posts.ID = $wpdb->term_relationships.object_id) "
                            . "INNER JOIN $wpdb->postmeta ON ($wpdb->posts.ID = $wpdb->postmeta.post_id) "
                            . "WHERE $wpdb->posts.post_type = 'product_variation' "
                            . "AND $wpdb->posts.post_status IN ('publish', 'private') "
                            . "AND ($wpdb->postmeta.meta_key = 'attribute_{$taxonomy}' "
                            . "AND $wpdb->postmeta.meta_value IN ("
                            . implode(',', array_map(
                                function ($value) {
                                    return "'" . esc_sql($value) . "'";
                                },
                                get_terms([
                                    'taxonomy' => $taxonomy,
                                    'include' => $include,
                                    'fields' => 'id=>slug'
                                ])
                            ))
                            . ")) ";

                        $results = $wpdb->get_results($queryString);
                        $productsIds = [];

                        foreach ($results as $result) {
                            $productsIds[$result->post_parent][] = (int) $result->term_taxonomy_id;
                        }

                        foreach ($productsIds as $id => $variations) {
                            $counts = array_count_values($variations);

                            if (isset($counts[$visibilityTerms['outofstock']])
                                && $counts[$visibilityTerms['outofstock']] == count($variations)
                            ) {
                                $postNotIn[] = $id;
                            }
                        }
                    }

                    if ($viewType == 'formula') {
                        // formula field type
                        $filterBy = [];
                        $terms = get_terms(['taxonomy' => $taxonomy]);
                        $valueSource = Question::getSetting($questionId, 'attribute_value_source');

                        foreach ($terms as $term) {
                            switch ($valueSource) {
                                default:
                                case 'name':
                                    $termValue = (float) $term->name;
                                    break;

                                case 'slug':
                                    $termValue = (float) $term->slug;
                            }

                            $met = false;

                            switch ($question->formulaOperator) {
                                case '=':
                                    $met = $termValue == $formulaValue;
                                    break;

                                case '!=':
                                    $met = $termValue != $formulaValue;
                                    break;

                                case '<':
                                    $met = $termValue < (float) $formulaValue;
                                    break;

                                case '<=':
                                    $met = $termValue <= (float) $formulaValue;
                                    break;

                                case '>':
                                    $met = $termValue > (float) $formulaValue;
                                    break;

                                case '>=':
                                    $met = $termValue >= (float) $formulaValue;
                            }

                            if ($met) {
                                $filterBy[] = $term->term_id;
                            }
                        }

                        $taxQuery[] = [
                            'taxonomy' => $taxonomy,
                            'field' => 'id',
                            'terms' => $filterBy,
                            'operator' => 'IN'
                        ];
                    } elseif (strpos($viewType, 'multi-choice') !== false) {
                        $in = [];

                        foreach (array_filter($questionValue) as $termId => $choiceValue) {
                            if ($choiceValue == 1) {
                                // get all IN values in the array
                                $in[] = $termId;

                                continue;
                            }

                            $taxQuery[] = [
                                'taxonomy' => $taxonomy,
                                'field' => 'id',
                                'terms' => $termId,
                                'operator' => $multiChoiceOperators[$choiceValue]
                            ];
                        }

                        if (!empty($in)) {
                            $taxQuery[] = [
                                'taxonomy' => $taxonomy,
                                'field' => 'id',
                                'terms' => $in,
                                'operator' => $multiChoiceOperators[1]
                            ];
                        }
                    } elseif (!empty(array_filter($questionValue))) {
                        // skip "doesn't matter" and not-defined value and filter by other terms
                        $taxQuery[] = [
                            'taxonomy' => $taxonomy,
                            'field' => 'id',
                            'terms' => array_filter($questionValue),
                            'operator' => $operator
                        ];
                    }
                }
            } elseif ($filterType == 'manual') {
                // filtering by a manual value
                foreach ($questionValue as $questionValueItem) {
                    if (!isset($question->manualValues[$questionValueItem])) {
                        continue;
                    }

                    $filterBy = $question->manualValues[$questionValueItem]['filter-by'];
                    $items = isset($question->manualValues[$questionValueItem]['items'])
                        ? $question->manualValues[$questionValueItem]['items']
                        : [];

                    $items = !empty($items) ? $items : [-1]; // if nothing is selected

                    if ($filterBy == 'any') {
                        continue;
                    } elseif ($filterBy == 'category') {
                        $taxQuery[] = [
                            'taxonomy' => 'product_cat',
                            'field' => 'id',
                            'terms' => array_filter($items, function ($item) {
                                return is_numeric($item);
                            }),
                            'operator' => $operator
                        ];
                    } elseif ($filterBy == 'tag') {
                        $taxQuery[] = [
                            'taxonomy' => 'product_tag',
                            'field' => 'id',
                            'terms' => array_filter($items, function ($item) {
                                return is_numeric($item);
                            }),
                            'operator' => $operator
                        ];
                    } elseif ($filterBy == 'attribute') {
                        $attributesQueries = [];
                        $index = 0;

                        foreach ($items as $item) {
                            if (empty($item['attribute']) || empty($item['attribute'])) {
                                continue;
                            }

                            if ($item['relation'] == 'and' && end($items) != $item) {
                                $attributesQueries[$index]['relation'] = 'AND';
                                $attributesQueries[$index][] = [
                                    'taxonomy' => 'pa_' . esc_sql($item['attribute']),
                                    'field' => 'id',
                                    'terms' => (int) $item['value']
                                ];
                            } else {
                                $attributesQueries[$index++][] = [
                                    'taxonomy' => 'pa_' . esc_sql($item['attribute']),
                                    'field' => 'id',
                                    'terms' => (int) $item['value']
                                ];
                            }
                        }

                        $taxQuery[] = array_merge(['relation' => 'or'], $attributesQueries);
                    } elseif ($filterBy == 'product') {
                        if (empty($postIn)) {
                            $postIn = $items;
                        } else {
                            $postIn = array_intersect($postIn, $items);
                        }
                    }
                }
            } elseif (in_array($filterType, ['price', 'meta'])) {
                // filtering by price or meta
                $metaKey = $filterType == 'price' ? '_price' : $question->metaKey;
                $metaValue = null;
                $compare = '=';
                $type = 'CHAR';

                if ($viewType == 'formula') {
                    $type = 'NUMERIC';
                    $compare = $question->formulaOperator;
                    $metaValue = $formulaValue;
                } elseif (strpos($viewType, 'number') !== false) {
                    $type = 'NUMERIC';

                    switch ($viewType) {
                        case 'number':
                            $compare = '=';
                            $metaValue = !empty($questionValue) ? reset($questionValue) : null;

                            break;
                        case 'number-from':
                            $compare = '>=';

                            if (isset($questionValue['from']) && !empty($questionValue['from'])) {
                                $metaValue = $questionValue['from'];
                            }

                            break;
                        case 'number-to':
                            $compare = '<=';

                            if (isset($questionValue['to']) && !empty($questionValue['to'])) {
                                $metaValue = $questionValue['to'];
                            }

                            break;
                        case 'number-between':
                            if (isset($questionValue['from']) && !empty($questionValue['from'])
                                && isset($questionValue['to'])
                                && !empty($questionValue['to'])
                            ) {
                                $compare = 'BETWEEN';
                                $metaValue = array_values($questionValue);
                            } elseif (isset($questionValue['from']) && !empty($questionValue['from'])) {
                                $compare = '>=';
                                $metaValue = $questionValue['from'];
                            } elseif (isset($questionValue['to']) && !empty($questionValue['to'])) {
                                $compare = '<=';
                                $metaValue = $questionValue['to'];
                            }
                    }
                } elseif (strpos($viewType, 'multi-choice') !== false) {
                    $in = [];

                    foreach (array_filter($questionValue) as $metaValue => $choiceValue) {
                        if ($choiceValue == 1) {
                            // get all IN values in the array
                            $in[] = $metaValue;

                            continue;
                        }

                        $metaQuery[] = [
                            'key' => $metaKey,
                            'value' => $metaValue,
                            'compare' => $multiChoiceOperators[$choiceValue],
                            'type' => $type
                        ];
                    }

                    if (!empty($in)) {
                        $metaQuery[] = [
                            'key' => $metaKey,
                            'value' => $in,
                            'compare' => $multiChoiceOperators[1],
                            'type' => $type
                        ];
                    }

                    $metaValue = null;
                } else {
                    $compare = 'IN';
                    $metaValue = $questionValue;
                }

                if (!is_null($metaValue)) {
                    $metaQuery[] = [
                        'key' => $metaKey,
                        'value' => $metaValue,
                        'compare' => $compare,
                        'type' => $type
                    ];
                }
            }
        }

        return apply_filters(
            'wcsf_query_args_by_filter',
            [
                'tax_query' => $taxQuery,
                'meta_query' => $metaQuery,
                'post__in' => $postIn,
                'post__not_in' => $postNotIn
            ],
            $id,
            $filterValue,
            $args
        );
    }

    /**
     * Filter the products on the shop page
     *
     * @param \WP_Query $query
     *
     * @return null
     */
    public function modifyProductsQuery($query)
    {
        if (is_admin() || !isset($_GET['wcsf-results']) || !isset($_GET['wcsf-value'])) {
            return;
        }

        do_action('wcsf_before_modify_request', $query);

        $id = (int) $_GET['wcsf-results'];
        $value = (array) $_GET['wcsf-value'];
        $args = $this->getQueryArgsByFilter($id, $value);

        if (method_exists($query, 'set')) {
            if (!empty($args['tax_query'])) {
                $query->set('tax_query', array_merge($query->get('tax_query'), $args['tax_query']));
            }

            if (!empty($args['post__in'])) {
                $query->set('post__in', array_merge($query->get('post__in'), $args['post__in']));
            }

            if (!empty($args['post__not_in'])) {
                $query->set('post__not_in', array_merge($query->get('post__not_in'), $args['post__not_in']));
            }

            if (!empty($args['meta_query'])) {
                $query->set('meta_query', array_merge($query->get('meta_query'), $args['meta_query']));
            }

            $maxResults = $this->filter->getSetting($id, 'max_results_number');
            $onlyResultRedirect = $this->filter->getSetting($id, 'redirect_to_only_result_page');
            $noResultsRedirect = $this->filter->getSetting($id, 'empty_results_url');

            if ($maxResults || $onlyResultRedirect || $noResultsRedirect) {
                add_filter(
                    'found_posts',
                    function ($found, $instance) use ($id, $maxResults, $onlyResultRedirect, $noResultsRedirect) {
                        if ($instance->query_vars['post_type'] != 'product') {
                            return $found;
                        }

                        // override products number in the query
                        if ($maxResults && property_exists($instance, 'posts') && is_array($instance->posts)) {
                            $limit = $maxResults;
                            $instance->post_count = $maxResults;
                            $instance->found_posts = $maxResults;

                            if ($instance->query_vars['posts_per_page'] < $maxResults) {
                                $limit = abs($instance->query_vars['paged'] * $instance->query_vars['posts_per_page']
                                    - $maxResults - $instance->query_vars['posts_per_page']);
                            }

                            $instance->posts = array_slice($instance->posts, 0, $limit);
                            $found = min($found, $maxResults);
                        }

                        if (($onlyResultRedirect || $noResultsRedirect) && in_array($found, [0, 1])) {
                            $redirectLink = null;

                            if ($onlyResultRedirect && (int) $found == 1) {
                                $redirectLink = get_permalink(reset($instance->posts));
                            } elseif ($noResultsRedirect && (int) $found == 0) {
                                $redirectLink = $noResultsRedirect;

                                // if URL is absolute
                                if (strpos($redirectLink, $_SERVER['HTTP_HOST']) === false) {
                                    $redirectLink = home_url() . '/' . $redirectLink;
                                }
                            }

                            $redirectLink = apply_filters('wcsf_results_redirect_link', $redirectLink, $id);
                            $redirectCode = apply_filters('wcsf_results_redirect_code', 302, $id, $redirectLink);

                            wp_redirect($redirectLink, $redirectCode);

                            // js version of redirect
                            exit("<script>document.location = '$redirectLink';</script>");
                        }

                        return $found;
                    },
                    15,
                    2
                );
            }
        }

        do_action('wcsf_after_modify_request', $query);

        // remove filter
        remove_action('woocommerce_product_query', [$this, 'modifyProductsQuery']);

        return;
    }

    /**
     * Filter woocommerce widget filter links
     *
     * @param string $link
     *
     * @return string
     */
    public function layeredNavLinkFilter($link)
    {
        $keys = [
            'wcsf-results',
            'wcsf-value',
            'wcsf-referer'
        ];

        foreach ($keys as $key) {
            if (isset($_GET[$key])) {
                $link = add_query_arg($key, $_GET[$key], $link);
            }
        }

        return $link;
    }

    /**
     * Filter the WOOF plugin query
     *
     * @param array $query
     *
     * @return array
     */
    public function modifyWoofQuery($query)
    {
        if (is_admin() || !isset($_GET['wcsf-results']) || !isset($_GET['wcsf-value'])) {
            return $query;
        }

        do_action('wcsf_before_modify_request', $query);

        $id = (int) $_GET['wcsf-results'];
        $value = (array) $_GET['wcsf-value'];
        $args = $this->getQueryArgsByFilter($id, $value);
        $maxResultsNumber = $this->filter->getSetting($id, 'max_results_number');
        $query = array_replace($query, $args);

        // override products number in the query
        if ($maxResultsNumber) {
            $query['posts_per_page'] = $maxResultsNumber;

            add_filter(
                'found_posts',
                function ($found) use ($maxResultsNumber) {
                    return min($found, $maxResultsNumber);
                }
            );
        }

        do_action('wcsf_after_modify_request', $query);

        // remove filter
        remove_action('woof_products_query', [$this, 'modifyWoofQuery']);

        return $query;
    }
    // </editor-fold>

    // <editor-fold desc="Assets">
    /** Admin styles and scripts enqueue */
    public function enqueueAdmin()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';
        $stylesFolder = WC_STEP_FILTER_DEBUG ? 'scss' : 'css';

        wp_enqueue_media();
        wp_enqueue_script('jquery');
        wp_enqueue_script('jquery-ui-sortable');
        wp_enqueue_script(
            'woocommerce-step-filter-data-table',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/data-table$suffix.js",
            [
                'jquery',
                'jquery-ui-sortable'
            ],
            WC_STEP_FILTER_VERSION,
            true
        );

        wp_enqueue_script(
            'woocommerce-step-filter-ajax-select',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/ajax-select$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );
        
        wp_enqueue_script(
            'woocommerce-step-filter-thumbnail',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/thumbnail$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );

        wp_enqueue_script(
            'woocommerce-step-filter-multi-select',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/multi-select$suffix.js",
            'jquery',
            WC_STEP_FILTER_PLUGIN_URL,
            true
        );

        wp_enqueue_script(
            'woocommerce-step-filter-hooks',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/hooks$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );

        wp_enqueue_style(
            'woocommerce-step-filter',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/$stylesFolder/app$suffix.css",
            [],
            WC_STEP_FILTER_VERSION
        );
    }

    /** Styles and scripts enqueue */
    public function enqueueFront()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';
        $stylesFolder = WC_STEP_FILTER_DEBUG ? 'scss' : 'css';
        $scriptsIncludingType = self::getSetting('scripts_including_type');
        $includedScripts = self::getSetting('included_scripts');
        $stylesIncludingType = self::getSetting('styles_including_type');

        wp_enqueue_script('jquery');

        if ($scriptsIncludingType == 'single' && !WC_STEP_FILTER_DEBUG) {
            wp_enqueue_script(
                'woocommerce-step-filter-scripts',
                WC_STEP_FILTER_PLUGIN_URL . 'assets/front/js/scripts.min.js',
                ['jquery'],
                WC_STEP_FILTER_VERSION,
                true
            );
        } elseif ($scriptsIncludingType == 'multiple' || WC_STEP_FILTER_DEBUG) {
            if (in_array('wNumb', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-wNumb',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/wNumb$suffix.js",
                    ['jquery'],
                    '1.1.0',
                    true
                );
            }

            if (in_array('nouislider', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-nouislider',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/nouislider$suffix.js",
                    ['jquery', 'woocommerce-step-filter-wNumb'],
                    '14.6.3',
                    true
                );
            }

            if (in_array('nouislider-launcher', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-nouislider-launcher',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/nouislider-launcher$suffix.js",
                    ['jquery', 'woocommerce-step-filter-nouislider'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }

            if (in_array('app', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-app',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/app$suffix.js",
                    ['jquery'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }

            if (in_array('hooks', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-hooks',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/hooks$suffix.js",
                    ['jquery', 'woocommerce-step-filter-app'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }
        }

        switch ($stylesIncludingType) {
            case 'full':
                wp_enqueue_style(
                    'woocommerce-step-filter-full',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/$stylesFolder/app-full$suffix.css",
                    [],
                    WC_STEP_FILTER_VERSION
                );

                break;
            case 'basic':
                wp_enqueue_style(
                    'woocommerce-step-filter',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/$stylesFolder/app$suffix.css",
                    [],
                    WC_STEP_FILTER_VERSION
                );
        }
    }
    // </editor-fold>

    // <editor-fold desc="Settings">
    /**
     * Add the tab to the woocommerce options page
     *
     * @param array $settingsTabs
     *
     * @return array $settingsTabs
     */
    public function addSettingsTab($settingsTabs)
    {
        $settingsTabs['step_filter_settings_tab'] = 'Step Filter';

        return $settingsTabs;
    }

    /** Init options fields */
    public function initSettingsTab()
    {
        $defaults = [
            'section_title' => [
                'name' => 'Step Filter',
                'type' => 'title',
                'desc' => '',
                'id' => 'wc_step_filter_section_title'
            ]
        ];

        $args = array_merge($defaults, self::$settingsModel);

        woocommerce_admin_fields($args);
        woocommerce_admin_fields(
            [
                'section_end' => [
                    'type' => 'sectionend',
                    'id' => 'wc_step_filter_section_end'
                ]
            ]
        );
    }

    /** Save options fields */
    public function updateSettingsTab()
    {
        woocommerce_update_options(self::$settingsModel);
    }

    /**
     * Get one of settings
     *
     * @param string $setting
     *
     * @return string|boolean|array|float
     */
    public static function getSetting($setting)
    {
        if (!isset(self::$settingsModel[$setting])) {
            return apply_filters('wcsf_global_setting', null, $setting);
        }

        static $globalSettingsCache;

        if (isset($globalSettingsCache[$setting])) {
            return apply_filters('wcsf_global_setting', $globalSettingsCache[$setting], $setting);
        }

        $value = get_option(self::$settingsModel[$setting]['key'], null);

        if (is_null($value) && isset(self::$settingsModel[$setting]['default'])) {
            $value = self::handleSettingType(
                self::$settingsModel[$setting]['default'],
                self::$settingsModel[$setting]['type']
            );
        }

        $globalSettingsCache[$setting] = self::handleSettingType($value, self::$settingsModel[$setting]['type']);

        return apply_filters('wcsf_global_setting', $globalSettingsCache[$setting], $setting);
    }

    /**
     * Generate html-field from passed args
     *
     * @param array $modelItem
     * @param array $args
     */
    public static function settingFieldView($modelItem, $args = [])
    {
        $defaultArgs = [
            'values' => [],
            'namePattern' => '%key%',
            'generateId' => true,
            'asTemplate' => false
        ];

        $args = array_replace_recursive($defaultArgs, $args);

        // create name from pattern
        $args['name'] = str_replace('%key%', $modelItem['key'], $args['namePattern']);

        // define value
        $args['value'] = isset($args['values'][$modelItem['key']])
            ? $args['values'][$modelItem['key']]
            : $modelItem['default'];

        // define id attribute
        $args['id'] = $args['generateId'] ? $modelItem['key'] : false;

        $viewPath = __DIR__ . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR
            . 'setting-field-views' . DIRECTORY_SEPARATOR . $modelItem['type'] . '.php';

        echo isset($modelItem['before']) ? $modelItem['before'] : '';

        if (file_exists($viewPath)) {
            include($viewPath);
        }

        echo isset($modelItem['after']) ? $modelItem['after'] : '';

        if (isset($modelItem['description'])) {
            echo '<p class="description">' . wp_kses_post($modelItem['description']) . '</p>';
        }
    }

    /**
     * Handle the setting value according to the type
     *
     * @param mixed $value
     * @param string $type
     *
     * @return string|float|bool|array
     */
    public static function handleSettingType($value, $type = 'string')
    {
        switch ($type) {
            case 'checkbox':
                $value = filter_var($value, FILTER_VALIDATE_BOOLEAN);
                break;

            case 'number':
                $value = (float) $value;
                break;

            case 'array':
                $value = (array) $value;
        }

        return $value;
    }
    // </editor-fold>

    // <editor-fold desc="Attribute">
    /**
     * Add metabox to attribute value edit page
     *
     * @param object $term
     */
    public function attributeValueFieldsAdd($term)
    {
        foreach (self::$attributeValueSettingsModel as $setting) {
            $value = get_term_meta($term->term_id, $setting['key']);
            ?>
            <tr valign="top">
                <th scope="row" class="titledesc">
                    <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                    echo wp_kses_post($setting['label']);
                    ?></label>
                </th>
                <td><?php
                    self::settingFieldView(
                        $setting,
                        ['values' => [$setting['key'] => isset($value[0]) ? $value[0] : null]]
                    );
                    ?></td>
            </tr>
            <?php
        }
    }

    /**
     * Save attribute value meta fields
     *
     * @param integer $termId
     */
    public function attributeValueFieldsSave($termId)
    {
        if (!$termId) {
            return;
        }

        foreach (self::$attributeValueSettingsModel as $setting) {
            if (!isset($_POST[$setting['key']])) {
                continue;
            }

            update_term_meta($termId, $setting['key'], $_POST[$setting['key']]);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Category">
    /**
     * Add metabox to category edit page
     *
     * @param object $term
     */
    public function categoryFieldsAdd($term)
    {
        foreach (self::$categorySettingsModel as $setting) {
            $value = get_term_meta($term->term_id, $setting['key']);
            ?>
            <tr valign="top">
                <th scope="row" class="titledesc">
                    <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                        echo wp_kses_post($setting['label']);
                        ?></label>
                </th>
                <td><?php
                    self::settingFieldView(
                        $setting,
                        ['values' => [$setting['key'] => isset($value[0]) ? $value[0] : null]]
                    );
                    ?></td>
            </tr>
            <?php
        }
    }

    /**
     * Save tag meta fields
     *
     * @param integer $termId
     */
    public function categoryFieldsSave($termId)
    {
        if (!$termId) {
            return;
        }

        foreach (self::$categorySettingsModel as $setting) {
            if (!isset($_POST[$setting['key']])) {
                continue;
            }

            update_term_meta($termId, $setting['key'], $_POST[$setting['key']]);
        }
    }
    // </editor-fold>

    // <editor-fold desc="AJAX">
    /** Search for categories and echo json */
    public function jsonSearchCategories()
    {
        ob_start();
        check_ajax_referer('search-products', 'security');

        $args = [
            'taxonomy' => 'product_cat',
            'orderby' => 'id',
            'order' => 'ASC',
            'hide_empty' => false,
            'include' => !empty($_GET['include']) ? wp_parse_id_list($_GET['include']) : '',
            'exclude' => !empty($_GET['exclude']) ? wp_parse_id_list($_GET['exclude']) : '',
            'number' => !empty($_GET['limit']) ? intval($_GET['limit']) : '',
            'name__like' => $_GET['term']
        ];

        $terms = get_terms($args);
        $foundTerms = [];

        if (!empty($terms)) {
            foreach ($terms as $term) {
                $foundTerms[$term->term_id] = rawurldecode($term->name);
            }
        }

        Utils::sendJSON(apply_filters('woocommerce_json_search_found_categories', $foundTerms));
    }

    /** Search for tags and echo json */
    public function jsonSearchTags()
    {
        ob_start();
        check_ajax_referer('search-products', 'security');

        $args = [
            'taxonomy' => 'product_tag',
            'orderby' => 'id',
            'order' => 'ASC',
            'hide_empty' => false,
            'include' => !empty($_GET['include']) ? wp_parse_id_list($_GET['include']) : '',
            'exclude' => !empty($_GET['exclude']) ? wp_parse_id_list($_GET['exclude']) : '',
            'number' => !empty($_GET['limit']) ? intval($_GET['limit']) : '',
            'name__like' => $_GET['term']
        ];

        $terms = get_terms($args);
        $foundTerms = [];

        if (!empty($terms)) {
            foreach ($terms as $term) {
                $foundTerms[$term->term_id] = rawurldecode($term->name);
            }
        }

        Utils::sendJSON(apply_filters('woocommerce_json_search_found_tags', $foundTerms));
    }

    /** Get list of an attribute values via Ajax */
    public function getAttributeValuesSelectOptionsAjax()
    {
        $value = $_GET['value'];

        if (!$value) {
            Utils::sendJSON();
        }

        $values = get_terms([
            'taxonomy' => "pa_{$value}",
            'hide_empty' => false,
            'fields' => 'id=>name'
        ]);

        Utils::sendJSON($values);
    }
    // </editor-fold>
}
