<?php
namespace WCStepFilter;

/**
 * Filter Class
 *
 * @class Filter
 * @version 5.10.0
 */
class Filter
{
    // <editor-fold desc="Properties">
    /**
     * Settings model variable
     *
     * @var array
     */
    public static $settingsModel;

    /**
     * Active steps session keys variable
     *
     * @var string
     */
    public static $activeStepsSessionKey = 'wcsf-active-step';

    /**
     * Filter value session keys variable
     *
     * @var string
     */
    public static $valuesSessionKey = 'wcsf-values';

    /**
     * Notices array
     * @var array
     */
    public $notices = [];
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Constructor */
    public function __construct()
    {
        add_action('init', [$this, 'onInit'], 5);
        add_action('woocommerce_after_register_post_type', [$this, 'wcProductRegistered'], 1);

        // admin scripts enqueue
        add_action('admin_enqueue_scripts', [$this, 'enqueueAdmin']);

        // post fields
        add_action('add_meta_boxes', [$this, 'addMetaBoxes']);
        add_action('save_post_wcsf', [$this, 'savePostAction']);

        // admin columns
        add_filter('manage_wcsf_posts_columns', [$this, 'postColumnsFilter']);
        add_action('manage_wcsf_posts_custom_column', [$this, 'postCustomColumnFilter'], 10, 2);

        // ajax
        add_action('wp_ajax_nopriv_WCSFGetStep', [$this, 'getStepHtmlAjax']);
        add_action('wp_ajax_WCSFGetStep', [$this, 'getStepHtmlAjax']);
        add_action('wp_ajax_nopriv_WCSFSubmit', [$this, 'submitAjax']);
        add_action('wp_ajax_WCSFSubmit', [$this, 'submitAjax']);
        add_action('wp_ajax_nopriv_WCSFSubmitAndGoTo', [$this, 'submitAndGoToAjax']);
        add_action('wp_ajax_WCSFSubmitAndGoTo', [$this, 'submitAndGoToAjax']);
        add_action('wp_ajax_nopriv_WCSFSubmitAndGoNext', [$this, 'submitAndGoNextAjax']);
        add_action('wp_ajax_WCSFSubmitAndGoNext', [$this, 'submitAndGoNextAjax']);
        add_action('wp_ajax_nopriv_WCSFSubmitAndGoResults', [$this, 'submitAndGoResultsAjax']);
        add_action('wp_ajax_WCSFSubmitAndGoResults', [$this, 'submitAndGoResultsAjax']);
        add_action('wp_ajax_nopriv_WCSFReset', [$this, 'resetAjax']);
        add_action('wp_ajax_WCSFReset', [$this, 'resetAjax']);
        add_action('wp_ajax_nopriv_WCSFGetPreliminaryResultsCount', [$this, 'getPreliminaryResultsCountAjax']);
        add_action('wp_ajax_WCSFGetPreliminaryResultsCount', [$this, 'getPreliminaryResultsCountAjax']);
    }

    /** On init action */
    public function onInit()
    {
        $defaultResultsURL = '/';
        $controlsClassDescription = L10N::r('For more info see the')
            . ' <a href="' . WC_STEP_FILTER_PLUGIN_URL
            . 'documentation/index.html#filter-settings-controls" target="_blank">'
            . L10N::r('documentation') . '</a>';

        if (function_exists('wc_get_page_id')) {
            $defaultResultsURL = str_replace(home_url(), '', get_permalink(wc_get_page_id('shop')));
        }

        self::$settingsModel = apply_filters(
            'wcsf_settings_model',
            [
                // <editor-fold desc="Basic">
                'questions' => [
                    'label' => L10N::r('Select questions for using'),
                    'key' => '_questions',
                    'type' => 'multi-select',
                    'default' => [],
                    'values' => Question::getPostsIdsList(),
                    'group' => L10N::r('Basic')
                ],
                'questions_categories' => [
                    'label' => L10N::r('Select questions categories for using'),
                    'key' => '_questions_categories',
                    'type' => 'multi-select',
                    'default' => [],
                    'values' => [],
                    'group' => L10N::r('Basic')
                ],
                'source' => [
                    'label' => L10N::r('Source of the steps'),
                    'key' => '_source',
                    'type' => 'select',
                    'default' => 'posts',
                    'values' => [
                        'posts' => L10N::r('Posts'),
                        'categories' => L10N::r('Categories')
                    ],
                    'group' => L10N::r('Basic')
                ],
                'mode' => [
                    'label' => L10N::r('Mode'),
                    'key' => '_mode',
                    'type' => 'select',
                    'default' => 'free-walk',
                    'values' => [
                        'free-walk' => L10N::r('Free walk'),
                        'step-by-step' => L10N::r('Step by step'),
                        'single-step' => L10N::r('Single step'),
                        'sequence' => L10N::r('Sequence')
                    ],
                    'group' => L10N::r('Basic')
                ],
                'extra_query_args' => [
                    'label' => L10N::r('Extra query arguments'),
                    'type' => 'data-table',
                    'inModal' => true,
                    'key' => '_extra_query_args',
                    'default' => [],
                    'description' => L10N::r('Additional query arguments for filtering'),
                    'values' => [
                        'source' => [
                            'label' => L10N::r('Source'),
                            'key' => 'source',
                            'type' => 'select',
                            'default' => 'product',
                            'values' => [
                                'product' => L10n::r('Product'),
                                'category' => L10n::r('Category'),
                                'attribute' => L10n::r('Attribute')
                            ]
                        ],
                        'product' => [
                            'label' => L10N::r('Product'),
                            'key' => 'product',
                            'type' => 'wc-product-search',
                            'multiple' => false,
                            'default' => []
                        ],
                        'category' => [
                            'label' => L10N::r('Category'),
                            'key' => 'category',
                            'type' => 'wc-terms-search',
                            'multiple' => false,
                            'default' => []
                        ],
                        'attribute_name' => [
                            'label' => L10N::r('Attribute name'),
                            'key' => 'attribute_name',
                            'type' => 'select',
                            'default' => '',
                            'values' => []
                        ],
                        'attribute_value' => [
                            'label' => L10N::r('Attribute value'),
                            'key' => 'attribute_value',
                            'type' => 'ajax-select',
                            'target-parent' => '[data-component="wcsf-data-table-item"]',
                            'target-selector' => '[data-key="attribute_name"] select',
                            'action' => 'wcsfGetAttributeValuesSelectOptions',
                            'default' => '',
                        ],
                        'condition' => [
                            'label' => L10N::r('Condition'),
                            'key' => 'condition',
                            'type' => 'select',
                            'default' => '=',
                            'values' => [
                                '=' => '=',
                                '!=' => '!='
                            ]
                        ]
                    ],
                    'group' => L10N::r('Basic')
                ],
                // </editor-fold>
                // <editor-fold desc="Results">
                'results_url' => [
                    'label' => L10N::r('Results URL'),
                    'key' => '_results_url',
                    'type' => 'text',
                    'default' => $defaultResultsURL,
                    // phpcs:disable
                    'description' => L10N::r('Set # to stay on the same page (for example, if you are using the filter as a widget)'),
                    // phpcs:enable
                    'group' => L10N::r('Results')
                ],
                'empty_results_url' => [
                    'label' => L10N::r('Empty results URL'),
                    'key' => '_empty_results_url',
                    'type' => 'text',
                    'default' => '',
                    'description' => L10N::r('Open another page in a case of empty results'),
                    'group' => L10N::r('Results')
                ],
                'show_results_within' => [
                    'label' => L10N::r('Show results within the filter'),
                    'key' => '_show_results_within',
                    'type' => 'checkbox',
                    'default' => false,
                    'group' => L10N::r('Results')
                ],
                'results_tab_title' => [
                    'label' => L10N::r('Results tab title'),
                    'key' => '_results_tab_title',
                    'default' => L10N::r('Results'),
                    'type' => 'text',
                    'group' => L10N::r('Results')
                ],
                'results_tab_thumbnail' => [
                    'label' => L10N::r('Results tab thumbnail'),
                    'key' => '_results_tab_thumbnail',
                    'default' => '',
                    'type' => 'thumbnail',
                    'group' => L10N::r('Results')
                ],
                'show_preliminary_results_count' => [
                    'label' => L10N::r('Show preliminary results count'),
                    'key' => '_show_preliminary_results_count',
                    'type' => 'checkbox',
                    'default' => false,
                    'group' => L10N::r('Results')
                ],
                'preliminary_results_count_text' => [
                    'label' => L10N::r('Preliminary results count text'),
                    'key' => '_preliminary_results_count_text',
                    'default' => L10N::r('Products found: %d'),
                    'description' => L10N::r('%d - products count variable'),
                    'type' => 'text',
                    'group' => L10N::r('Results')
                ],
                'max_results_number' => [
                    'label' => L10N::r('Maximum products number in results'),
                    'key' => '_max_results_number',
                    'type' => 'number',
                    'min' => 0,
                    'default' => 0,
                    'description' => L10N::r('Zero is equal infinity'),
                    'group' => L10N::r('Results')
                ],
                'redirect_to_only_result_page' => [
                    'label' => L10N::r('Instant redirect to the only result page'),
                    'key' => '_redirect_to_only_result_page',
                    'type' => 'checkbox',
                    'default' => false,
                    'description' => L10N::r('Not works for within results showing'),
                    'group' => L10N::r('Results')
                ],
                // </editor-fold>
                // <editor-fold desc="Behavior">
                'collect_inquiries' => [
                    'label' => L10N::r('Collect inquiries for statistic'),
                    'key' => '_collect_inquiries',
                    'type' => 'checkbox',
                    'default' => false,
                    'group' => L10N::r('Behavior')
                ],
                'scrolling_top_on_update' => [
                    'label' => L10N::r('Scrolling to top on the form update'),
                    'key' => '_scrolling_top_on_update',
                    'type' => 'checkbox',
                    'default' => true,
                    'public' => true,
                    'group' => L10N::r('Behavior')
                ],
                'scrolling_up_gap' => [
                    'label' => L10N::r('The gap on scrolling up (px)'),
                    'key' => '_scrolling_up_gap',
                    'type' => 'number',
                    'default' => 0,
                    'public' => true,
                    'group' => L10N::r('Behavior')
                ],
                'nav_action' => [
                    'label' => L10N::r('Nav action'),
                    'key' => '_nav_action',
                    'type' => 'select',
                    'default' => 'submit-and-go-to',
                    'values' => [
                        'submit-and-go-to' => L10N::r('Submit and go to step'),
                        'get-step' => L10N::r('Go to step')
                    ],
                    'group' => L10N::r('Behavior')
                ],
                // </editor-fold>
                // <editor-fold desc="View">
                'show_header' => [
                    'label' => L10N::r('Show header'),
                    'key' => '_show_header',
                    'type' => 'checkbox',
                    'default' => false,
                    'group' => L10N::r('View')
                ],
                'show_footer' => [
                    'label' => L10N::r('Show footer'),
                    'key' => '_show_footer',
                    'type' => 'checkbox',
                    'default' => true,
                    'group' => L10N::r('View')
                ],
                'nav_view' => [
                    'label' => L10N::r('Nav view'),
                    'key' => '_nav_view',
                    'type' => 'select',
                    'default' => 'buttons',
                    'values' => Core::getNavViewsList(),
                    'group' => L10N::r('View')
                ],
                'enable_description_tab' => [
                    'label' => L10N::r('Enable description tab'),
                    'key' => '_enable_description_tab',
                    'default' => true,
                    'type' => 'checkbox',
                    'group' => L10N::r('View')
                ],
                'description_tab_title' => [
                    'label' => L10N::r('Description tab title'),
                    'key' => '_description_tab_title',
                    'default' => L10N::r('Welcome'),
                    'type' => 'text',
                    'group' => L10N::r('View')
                ],
                'enable_value' => [
                    'label' => L10N::r('Enable value list'),
                    'key' => '_enable_value',
                    'type' => 'checkbox',
                    'default' => true,
                    'group' => L10N::r('View')
                ],
                'enable_shop_value' => [
                    'label' => L10N::r('Enable shop value list'),
                    'key' => '_enable_shop_value',
                    'default' => true,
                    'type' => 'checkbox',
                    'description' => L10N::r('Appears on the shop page while results showing'),
                    'group' => L10N::r('View')
                ],
                // </editor-fold>
                // <editor-fold desc="Results PDF">
                'results_pdf_header_height' => [
                    'label' => L10N::r('Results PDF header height'),
                    'key' => '_results_pdf_header_height',
                    'type' => 'number',
                    'min' => 0,
                    'default' => 2,
                    'description' => L10N::r('cm'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_header_content' => [
                    'label' => L10N::r('Results PDF header content'),
                    'key' => '_results_pdf_header_content',
                    'type' => 'editor',
                    'default' => '<div style="padding:2em;background-color:#f5f5f5">' . get_bloginfo('name') . '</div>',
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_top_description' => [
                    'label' => L10N::r('Results PDF top description'),
                    'key' => '_results_pdf_top_description',
                    'type' => 'editor',
                    'default' => L10N::r('Your filtering results'),
                    'description' =>
                        L10N::r('Allowed shortcodes: [wcsf-result-pdf-page-number], [wcsf-result-pdf-page-total]'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_footer_height' => [
                    'label' => L10N::r('Results PDF footer height'),
                    'key' => '_results_pdf_footer_height',
                    'type' => 'number',
                    'min' => 0,
                    'default' => 2,
                    'description' => L10N::r('cm'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_footer_content' => [
                    'label' => L10N::r('Results PDF footer content'),
                    'key' => '_results_pdf_footer_content',
                    'type' => 'editor',
                    'default' => '<div style="padding:2em;background-color:#f5f5f5;">'
                        . '[wcsf-result-pdf-page-number] / [wcsf-result-pdf-page-total]'
                        . '</div>',
                    'description' =>
                        L10N::r('Allowed shortcodes: [wcsf-result-pdf-page-number], [wcsf-result-pdf-page-total]'),
                    'group' => L10N::r('Results PDF')
                ],
                'results_pdf_bottom_description' => [
                    'label' => L10N::r('Results PDF bottom description'),
                    'key' => '_results_pdf_bottom_description',
                    'type' => 'editor',
                    'default' => 'Get more info: ' . get_bloginfo('url'),
                    'group' => L10N::r('Results PDF')
                ],
                // </editor-fold>
                // <editor-fold desc="Controls">
                'enable_back_button' => [
                    'label' => L10N::r('Enable "Back" button'),
                    'key' => '_enable_back_button',
                    'default' => true,
                    'type' => 'checkbox',
                    'group' => L10N::r('Controls')
                ],
                'back_button_text' => [
                    'label' => L10N::r('"Back" button text'),
                    'key' => '_back_button_text',
                    'default' => L10N::r('Back'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'back_button_class' => [
                    'label' => L10N::r('"Back" button class'),
                    'key' => '_back_button_class',
                    'type' => 'text',
                    'default' => 'btn-default btn-light show-icon-on-mobile icon-left hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'enable_skip_button' => [
                    'label' => L10N::r('Enable "Skip" button'),
                    'key' => '_enable_skip_button',
                    'default' => true,
                    'type' => 'checkbox',
                    'group' => L10N::r('Controls')
                ],
                'skip_button_text' => [
                    'label' => L10N::r('"Skip" button text'),
                    'key' => '_skip_button_text',
                    'default' => L10N::r('Skip'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'skip_button_class' => [
                    'label' => L10N::r('"Skip" button class'),
                    'key' => '_skip_button_class',
                    'type' => 'text',
                    'default' => 'btn-default btn-light show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'show_next_button' => [
                    'label' => L10N::r('Show "Next" button'),
                    'key' => '_show_next_button',
                    'default' => 'until-last-step',
                    'type' => 'select',
                    'values' => [
                        'until-last-step' => L10N::r('Until the last step'),
                        'always' => L10N::r('Always'),
                        'never' => L10N::r('Never')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'next_button_text' => [
                    'label' => L10N::r('"Next" button text'),
                    'key' => '_next_button_text',
                    'default' => L10N::r('Next'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'next_button_class' => [
                    'label' => L10N::r('"Next" button class'),
                    'key' => '_next_button_class',
                    'type' => 'text',
                    'default' => 'btn-primary show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'enable_reset_button' => [
                    'label' => L10N::r('Enable "Reset" button'),
                    'key' => '_enable_reset_button',
                    'default' => true,
                    'type' => 'checkbox',
                    'group' => L10N::r('Controls')
                ],
                'reset_button_text' => [
                    'label' => L10N::r('"Reset" button text'),
                    'key' => '_reset_button_text',
                    'default' => L10N::r('Reset'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'reset_button_class' => [
                    'label' => L10N::r('"Reset" button class'),
                    'key' => '_reset_button_class',
                    'type' => 'text',
                    'default' => 'btn-warning show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'show_results_button' => [
                    'label' => L10N::r('Show "Results" button'),
                    'key' => '_show_results_button',
                    'default' => 'on-answer',
                    'type' => 'select',
                    'values' => [
                        'on-answer' => L10N::r('Then have an answer'),
                        'last-step' => L10N::r('On the last step'),
                        'always' => L10N::r('Always'),
                        'never' => L10N::r('Never')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'results_button_text' => [
                    'label' => L10N::r('"Results" button text'),
                    'key' => '_results_button_text',
                    'default' => L10N::r('Show results'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'results_button_class' => [
                    'label' => L10N::r('"Results" button class'),
                    'key' => '_results_button_class',
                    'type' => 'text',
                    'default' => 'btn-success show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'show_results_pdf_button' => [
                    'label' => L10N::r('Show "Results PDF" button'),
                    'key' => '_show_results_pdf_button',
                    'default' => 'on-answer',
                    'type' => 'select',
                    'values' => [
                        'on-answer' => L10N::r('Then have an answer'),
                        'last-step' => L10N::r('On the last step'),
                        'always' => L10N::r('Always'),
                        'never' => L10N::r('Never')
                    ],
                    'group' => L10N::r('Controls')
                ],
                'results_pdf_button_text' => [
                    'label' => L10N::r('"Results PDF" button text'),
                    'key' => '_results_pdf_button_text',
                    'type' => 'text',
                    'default' => L10N::r('Get PDF'),
                    'group' => L10N::r('Controls')
                ],
                'results_pdf_button_class' => [
                    'label' => L10N::r('"Results PDF" button class'),
                    'key' => '_results_pdf_button_class',
                    'type' => 'text',
                    'default' => 'btn-info show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'enable_shop_reset_button' => [
                    'label' => L10N::r('Enable shop "Reset" button'),
                    'key' => '_enable_shop_reset_button',
                    'default' => true,
                    'type' => 'checkbox',
                    'group' => L10N::r('Controls')
                ],
                'shop_reset_button_text' => [
                    'label' => L10N::r('Shop "Reset" button text'),
                    'key' => '_shop_reset_button_text',
                    'default' => L10N::r('Reset'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'shop_reset_button_class' => [
                    'label' => L10N::r('Shop "Reset" button class'),
                    'key' => '_shop_reset_button_class',
                    'type' => 'text',
                    'default' => 'btn-warning show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'enable_shop_back_button' => [
                    'label' => L10N::r('Enable shop "Back" button'),
                    'key' => '_enable_shop_back_button',
                    'default' => true,
                    'type' => 'checkbox',
                    'group' => L10N::r('Controls')
                ],
                'shop_back_button_text' => [
                    'label' => L10N::r('Shop "Back" button text'),
                    'key' => '_shop_back_button_text',
                    'default' => L10N::r('Back'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'shop_back_button_url' => [
                    'label' => L10N::r('Shop "Back" button URL'),
                    'key' => '_shop_back_button_url',
                    'default' => '',
                    'type' => 'text',
                    'description' => L10N::r('Keep blank to get previous page automatically'),
                    'group' => L10N::r('Controls')
                ],
                'shop_back_button_class' => [
                    'label' => L10N::r('Shop "Back" button class'),
                    'key' => '_shop_back_button_class',
                    'type' => 'text',
                    'default' => 'btn-default btn-light show-icon-on-mobile icon-left hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ],
                'enable_shop_repeat_button' => [
                    'label' => L10N::r('Enable shop "Back" button'),
                    'key' => '_enable_shop_repeat_button',
                    'default' => true,
                    'type' => 'checkbox',
                    'group' => L10N::r('Controls')
                ],
                'shop_repeat_button_text' => [
                    'label' => L10N::r('Shop "Repeat" button text'),
                    'key' => '_shop_repeat_button_text',
                    'default' => L10N::r('Repeat'),
                    'type' => 'text',
                    'group' => L10N::r('Controls')
                ],
                'shop_repeat_button_url' => [
                    'label' => L10N::r('Shop "Repeat" button URL'),
                    'key' => '_shop_repeat_button_url',
                    'default' => '',
                    'type' => 'text',
                    'description' => L10N::r('Keep blank to get previous page automatically'),
                    'group' => L10N::r('Controls')
                ],
                'shop_repeat_button_class' => [
                    'label' => L10N::r('Shop "Repeat" button class'),
                    'key' => '_shop_repeat_button_class',
                    'type' => 'text',
                    'default' => 'btn-info show-icon-on-mobile icon-right hide-text-on-mobile',
                    'description' => $controlsClassDescription,
                    'group' => L10N::r('Controls')
                ]
                // </editor-fold>
            ]
        );

        $name = L10N::r('WooCommerce Step Filter');
        $args = [
            'label' => $name,
            'labels' => [
                'name' => $name,
                'singular_name' => $name,
                'menu_name' => L10N::r('WC Step Filter')
            ],
            'description' => L10N::r('This is where you can add new step filter items'),
            'public' => false,
            'show_ui' => true,
            'map_meta_cap' => true,
            'publicly_queryable' => false,
            'exclude_from_search' => true,
            'show_in_menu' => true,
            'hierarchical' => false,
            'rewrite' => false,
            'query_var' => false,
            'menu_icon' => 'dashicons-search',
            'supports' => [
                'title',
                'thumbnail',
                'editor'
            ],
            'show_in_nav_menus' => false,
            'show_in_admin_bar' => true
        ];

        register_post_type('wcsf', $args);
    }

    /** Handles then woocommerce product is registered */
    public function wcProductRegistered()
    {
        if (isset($_GET['wcsf-results-pdf']) && $_GET['wcsf-results-pdf']) {
            $this->outputResultPDF((int) $_GET['wcsf-results-pdf']);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Admin">
    /** Admin styles and scripts enqueue */
    public function enqueueAdmin()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';

        wp_enqueue_script(
            'woocommerce-step-filter-filter',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/filter$suffix.js",
            'jquery',
            WC_STEP_FILTER_PLUGIN_URL,
            true
        );

        if (defined('WC_VERSION')
            && ((isset($_GET['post']) && get_post_type((int) $_GET['post']) == 'wcsf')
                || (isset($_GET['post_type']) && $_GET['post_type'] == 'wcsf'))
        ) {
            wp_register_script(
                'select2',
                \WC()->plugin_url() . "/assets/js/select2/select2.full$suffix.js",
                ['jquery'],
                '4.0.3'
            );

            wp_register_script(
                'selectWoo',
                WC()->plugin_url() . "/assets/js/selectWoo/selectWoo.full$suffix.js",
                ['jquery'],
                '1.0.0'
            );

            wp_register_script(
                'wc-enhanced-select',
                WC()->plugin_url() . "/assets/js/admin/wc-enhanced-select$suffix.js",
                ['jquery', 'selectWoo'],
                WC_VERSION
            );

            wp_enqueue_script('select2');
            wp_enqueue_script('selectWoo');
            wp_enqueue_script('wc-enhanced-select');
            wp_enqueue_style('woocommerce_admin_styles', \WC()->plugin_url() . '/assets/css/admin.css', [], WC_VERSION);
        }
    }

    /**
     * Save meta values
     *
     * @param integer $postId
     */
    public function savePostAction($postId)
    {
        if ((defined('DOING_AUTOSAVE') && DOING_AUTOSAVE)
            || !current_user_can('edit_page', $postId)
            || (isset($_REQUEST['action']) && $_REQUEST['action'] == 'inline-save')
        ) {
            return;
        }

        foreach (self::$settingsModel as $setting) {
            if (!isset($_POST[$setting['key']])) {
                continue;
            }

            update_post_meta($postId, $setting['key'], $_POST[$setting['key']]);
        }
    }

    /** Add meta-boxes */
    public function addMetaBoxes()
    {
        add_meta_box(
            'woocommerce-step-filter',
            L10N::r('Settings'),
            [$this, 'metaBoxCallback'],
            'wcsf',
            'normal'
        );
    }

    /**
     * Post page meta-box content
     *
     * @param object $post
     */
    public function metaBoxCallback($post)
    {
        $postSettingsModelGroups = [];
        $attributes = wc_get_attribute_taxonomies();
        $shortCodes = [
            'shortcode' => [
                'group' => L10N::r('Basic'),
                'label' => L10N::r('Shortcode for using'),
                'key' => '_shortcode',
                'type' => 'text',
                'readonly' => true,
                'default' => '[woocommerce-step-filter id="' . $post->ID . '"]'
            ]
        ];

        foreach ($attributes as $attribute) {
            self::$settingsModel['extra_query_args']['values']
                ['attribute_name']['values'][$attribute->attribute_name] = $attribute->attribute_label;
        }

        $attributeKeys = array_keys(self::$settingsModel['extra_query_args']['values']['attribute_name']['values']);
        self::$settingsModel['extra_query_args']['values']['attribute_name']['default'] = reset($attributeKeys);

        self::$settingsModel = $shortCodes + self::$settingsModel;
        self::$settingsModel['questions_categories']['values'] = Question::getCategoriesIdsList();

        foreach (self::$settingsModel as $settingKey => $setting) {
            $postSettingsModelGroups[$setting['group']][$settingKey] = $setting;
        }
        ?>
        <div data-component="wcsf-settings-groups">
            <section class="wcsf-filter wcsf-settings-group" data-component="wcsf-filter wcsf-settings-group">
                <?php foreach ($postSettingsModelGroups as $group => $settingsModel) { ?>
                    <fieldset>
                        <a class="button button-large wcsf-settings-group-toggle" href="#" role="button"
                            data-component="wcsf-settings-group-toggle"
                            data-id="<?php echo esc_attr($group); ?>"><?php echo wp_kses_post($group); ?></a>
                        <table class="form-table wcsf-settings-table wcsf-settings-group-content<?php
                        echo reset($postSettingsModelGroups) == $settingsModel ? ' is-visible' : '';
                        ?>"
                            data-component="wcsf-settings-group-content"
                            data-id="<?php echo esc_attr($group); ?>">
                            <?php foreach ($settingsModel as $settingKey => $setting) { ?>
                                <tr data-component="wcsf-setting" data-key="<?php echo esc_attr($settingKey); ?>">
                                    <th>
                                        <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                                            echo wp_kses_post($setting['label']);
                                            ?></label>
                                    </th>
                                    <td><?php
                                        Core::settingFieldView(
                                            $setting,
                                            ['values' => [$setting['key'] => self::getSetting($post->ID, $settingKey)]]
                                        );
                                        ?></td>
                                </tr>
                            <?php } ?>
                        </table>
                    </fieldset>
                <?php } ?>
            </section>
        </div>
        <?php
    }

    /**
     * Filters list columns filter
     *
     * @param array $columns
     *
     * @return array $columns
     */
    public function postColumnsFilter($columns)
    {
        $columns['shortcode'] = L10N::r('ShortCode');

        return $columns;
    }

    /**
     * Filters list line cell
     *
     * @param array $columns
     * @param integer $postId
     */
    public function postCustomColumnFilter($columns, $postId)
    {
        if ($columns == 'shortcode') {
            echo '[woocommerce-step-filter id="' . (int) $postId . '"]';
        }
    }
    // </editor-fold>

    // <editor-fold desc="Settings">
    /**
     * Get one of post settings
     *
     * @param integer $id
     * @param string $setting
     *
     * @return string|boolean|array|float
     */
    public static function getSetting($id, $setting)
    {
        static $filterSettingsCache;

        if (isset($filterSettingsCache[$id], $filterSettingsCache[$id][$setting])) {
            return apply_filters('wcsf_setting', $filterSettingsCache[$id][$setting], $id, $setting);
        }

        $value = isset(self::$settingsModel[$setting]['key'])
            ? get_post_meta($id, self::$settingsModel[$setting]['key'], true)
            : '';

        if ($value == '' && isset(self::$settingsModel[$setting]['default'])) {
            $value = self::$settingsModel[$setting]['default'];
        }

        $value = Core::handleSettingType($value, self::$settingsModel[$setting]['type']);

        if (!isset($filterSettingsCache[$id])) {
            $filterSettingsCache[$id] = [];
        }

        $filterSettingsCache[$id][$setting] = $value;

        return apply_filters('wcsf_setting', $value, $id, $setting);
    }

    /**
     * Get post settings
     *
     * @param integer $id
     * @param array $args
     *
     * @return array
     */
    public static function getSettings($id, $args = [])
    {
        $defaults = ['public' => false];
        $args = array_merge($defaults, $args);
        $output = [];

        foreach (self::$settingsModel as $settingModelKey => $settingModel) {
            if ($args['public'] && (!isset($settingModel['public']) || !$settingModel['public'])) {
                continue;
            }

            $output[$settingModelKey] = self::getSetting($id, $settingModelKey);
        }

        return $output;
    }

    /**
     * Get questions list setting for any source type
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getQuestionsList($id)
    {
        switch (self::getSetting($id, 'source')) {
            case 'categories':
                $output = [];
                $stepsIds = self::getSetting($id, 'questions_categories');

                foreach ($stepsIds as $stepId) {
                    $questions = Question::getPostsIdsList([
                        'tax_query' => [
                            [
                                'taxonomy' => 'wcsf_question_cat',
                                'terms' => $stepId
                            ]
                        ]
                    ]);

                    $output = array_replace(
                        $output,
                        Question::getList($id, array_keys($questions))
                    );
                }

                break;

            default:
            case 'posts':
                $output = Question::getList($id, self::getSetting($id, 'questions'));
        }

        return apply_filters('wcsf_questions_list', $output, $id);
    }
    // </editor-fold>

    // <editor-fold desc="Steps">
    /**
     * Get filter steps array
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getSteps($id)
    {
        static $stepsCache = [];

        if (isset($stepsCache[$id])) {
            return apply_filters('wcsf_steps', $stepsCache[$id], $id);
        }

        switch (self::getSetting($id, 'source')) {
            case 'categories':
                $steps = [];
                $stepsIds = self::getSetting($id, 'questions_categories');

                foreach ($stepsIds as $stepId) {
                    $term = get_term($stepId, 'wcsf_question_cat');
                    $questions = Question::getPostsIdsList([
                        'orderby' => 'menu_order',
                        'order' => 'ASC',
                        'tax_query' => [
                            [
                                'taxonomy' => 'wcsf_question_cat',
                                'terms' => $stepId
                            ]
                        ]
                    ]);

                    $steps[$stepId] = (object) [
                        'id' => $stepId,
                        'name' => $term->name,
                        'description' => $term->description,
                        'thumbnail' => get_term_meta($stepId, '_thumbnail', 1),
                        'children' => Question::getList($id, array_keys($questions))
                    ];
                }

                break;

            default:
            case 'posts':
                $steps = Question::getList($id, self::getSetting($id, 'questions'));
        }

        if (self::getSetting($id, 'enable_description_tab')) {
            $steps = array_replace(
                [
                    'description' => (object) [
                        'id' => 'description',
                        'name' => self::getSetting($id, 'description_tab_title'),
                        'description' => get_post_field('post_content', $id),
                        'thumbnail' => get_post_thumbnail_id($id)
                    ]
                ],
                $steps
            );
        }

        if (self::getSetting($id, 'show_results_within')) {
            $steps['results'] = (object) [
                'id' => 'results',
                'name' => self::getSetting($id, 'results_tab_title'),
                'thumbnail' => self::getSetting($id, 'results_tab_thumbnail')
            ];
        }

        $stepsCache[$id] = $steps;

        return apply_filters('wcsf_steps', $steps, $id);
    }

    /**
     * Get filter steps ids array
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getStepsIds($id)
    {
        $stepsIds = [];

        foreach (self::getSteps($id) as $step) {
            $stepsIds[] = $step->id;
        }

        return apply_filters('wcsf_steps_ids', $stepsIds, $id);
    }

    /**
     * Get filter active step id
     *
     * @param integer $id
     *
     * @return integer
     */
    public static function getActiveStepId($id)
    {
        $activeStep = Storage::get(self::$activeStepsSessionKey, $id);

        if (!$activeStep) {
            $stepsIds = self::getStepsIds($id);
            $activeStep = reset($stepsIds);
        }

        return $activeStep;
    }

    /**
     * Get filter active step
     *
     * @param integer $id
     *
     * @return object|null
     */
    public static function getActiveStep($id)
    {
        $steps = self::getSteps($id);
        $activeStepId = self::getActiveStepId($id);
        $metActiveItem = false;
        $mode = self::getSetting($id, 'mode');
        $output = null;

        foreach ($steps as $step) {
            if ($mode == 'step-by-step' && $metActiveItem) {
                // in the step-by-step mode reset all next steps after the active
                self::resetValue($id, $step->id);
            }

            if ($step->id == $activeStepId) {
                $metActiveItem = true;
                $output = $step;
            }
        }

        return apply_filters('wcsf_active_step', $output, $id);
    }

    /**
     * Get filter the next active step
     *
     * @param integer $id
     *
     * @return integer|string
     */
    public static function getNextActiveStepId($id)
    {
        $stepsIds = self::getStepsIds($id);
        $activeStepId = self::getActiveStepId($id);
        $foundActive = false;

        foreach ($stepsIds as $stepId) {
            if ($foundActive) {
                return $stepId;

                break;
            }

            if ($stepId == $activeStepId) {
                $foundActive = true;
            }
        }

        return null;
    }

    /**
     * Get filter the prev active step
     *
     * @param integer $id
     *
     * @return integer|string
     */
    public static function getPrevActiveStepId($id)
    {
        $stepsIds = self::getStepsIds($id);
        $activeStepId = self::getActiveStepId($id);
        $prevStep = null;

        foreach ($stepsIds as $stepId) {
            if ($stepId == $activeStepId) {
                return $prevStep;
            }

            $prevStep = $stepId;
        }

        return null;
    }

    /**
     * Set active filter step to the session variable
     *
     * @param integer $id
     * @param integer|string $step
     */
    public static function setActiveStep($id, $step)
    {
        Storage::set(self::$activeStepsSessionKey, $id, $step);
    }

    /**
     * Reset filter values and set the first step as active
     *
     * @param integer $id
     * @param integer $questionId
     */
    public static function resetState($id, $questionId = null)
    {
        self::resetValue($id, $questionId);

        if (!$questionId) {
            $stepsIds = self::getStepsIds($id);

            self::setActiveStep($id, $stepsIds[0]);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Notices">
    /**
     * Add notice by type into a queue
     *
     * @param string $stepId
     * @param array $massageData
     */
    public function addNotice($stepId, $massageData)
    {
        $this->notices[$stepId][] = $massageData;
    }

    /**
     * Return the queue of notices
     *
     * @param string $stepId - try to get messages from one step by id or output all of messages
     *
     * @return array
     */
    public function getNotices($stepId = null)
    {
        if ($stepId) {
            // return step's messages array or nothing
            return isset($this->notices[$stepId]) ? $this->notices[$stepId] : [];
        } else {
            // return all steps messages
            return array_reduce($this->notices, 'array_merge', []);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Value">
    /**
     * Save the filter value
     * Value should be an array like [questionId => [value1 => 123, value2 = 345]]
     *
     * @param integer $id
     * @param array $value
     *
     * @throws \Exception
     */
    public static function setValue($id, $value)
    {
        foreach ((array) $value as $questionId => $valueItem) {
            foreach ($valueItem as $stepValueKey => $stepValue) {
                // clear empty items, but not zero
                if ($stepValue == '') {
                    unset($valueItem[$stepValueKey]);
                }
            }

            $minValuesSelected = Question::getSetting($questionId, 'min_values_selected');
            $maxValuesSelected = Question::getSetting($questionId, 'max_values_selected');

            if ($minValuesSelected && count($valueItem) < $minValuesSelected) {
                $message = str_replace(
                    [
                        '%limit%',
                        '%value%'
                    ],
                    [
                        $minValuesSelected,
                        count($valueItem)
                    ],
                    Question::getSetting($questionId, 'min_values_selected_message')
                );

                throw new \Exception($message, $questionId);
            } elseif ($maxValuesSelected && count($valueItem) > $maxValuesSelected) {
                $message = str_replace(
                    [
                        '%limit%',
                        '%value%'
                    ],
                    [
                        $maxValuesSelected,
                        count($valueItem)
                    ],
                    Question::getSetting($questionId, 'max_values_selected_message')
                );

                throw new \Exception($message, $questionId);
            }

            Storage::set(self::$valuesSessionKey, $id, $valueItem, $questionId);
            $bondedQuestions = Question::getSetting($questionId, 'bonded_questions');

            // pass value to bonded questions
            if (!empty($bondedQuestions)) {
                $filtersIds = self::getPostsIdsList();

                foreach (array_keys($filtersIds) as $filterId) {
                    if ($filterId == $id) {
                        // skip the same filter
                        continue;
                    }

                    $questions = self::getQuestionsList($filterId);

                    foreach ($bondedQuestions as $bondedQuestion) {
                        if (!in_array($bondedQuestion, array_keys($questions))) {
                            continue;
                        }

                        Storage::set(self::$valuesSessionKey, $filterId, $valueItem, $bondedQuestion);
                    }
                }
            }
        }
    }

    /**
     * Reset the filter by step id or fully
     *
     * @param integer $id
     * @param integer $questionId
     */
    public static function resetValue($id, $questionId = null)
    {
        if ($questionId) {
            Storage::set(self::$valuesSessionKey, $id, [], $questionId);
            $bondedQuestions = Question::getSetting($questionId, 'bonded_questions');

            // pass value to bonded questions
            if (!empty($bondedQuestions)) {
                $filtersIds = self::getPostsIdsList();

                foreach (array_keys($filtersIds) as $filterId) {
                    if ($filterId == $id) {
                        // skip the same filter
                        continue;
                    }

                    $questions = self::getQuestionsList($filterId);

                    foreach ($bondedQuestions as $bondedQuestion) {
                        if (!in_array($bondedQuestion, array_keys($questions))) {
                            continue;
                        }

                        Storage::set(self::$valuesSessionKey, $filterId, [], $bondedQuestion);
                    }
                }
            }
        } else {
            Storage::set(self::$valuesSessionKey, $id, []);
        }
    }

    /**
     * Get the filter step value
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getValue($id)
    {
        $output = (array) Storage::get(self::$valuesSessionKey, $id);
        $output = self::removeBorderlineNumberValues($id, $output);

        return apply_filters('wcsf_value', array_filter($output), $id);
    }

    /**
     * Remove number view values which equals to their min/max values
     *
     * @param integer $id
     * @param array $values
     *
     * @return array
     */
    public static function removeBorderlineNumberValues($id, $values)
    {
        foreach (array_keys($values) as $questionId) {
            $question = Question::get($questionId);
            $boundary = [];

            if (!is_object($question) || !property_exists($question, 'viewType')
                || strpos($question->viewType, 'number') === false
            ) {
                continue;
            }

            // get min/max values
            switch ($question->filterType) {
                case 'price':
                    $boundary = Utils::getBoundaryPrices();
                    break;

                case 'attribute':
                    $boundary = Utils::getBoundaryAttributeValues($question->attribute);
                    break;

                case 'meta':
                    $boundary = Utils::getBoundaryMetaValues($question->metaKey);
            }

            switch ($question->viewType) {
                case 'number-from':
                    if (isset($values[$questionId]['from']) && $values[$questionId]['from'] == $boundary['min']) {
                        unset($values[$questionId]);
                    }

                    break;
                case 'number-to':
                    if (isset($values[$questionId]['to']) && $values[$questionId]['to'] == $boundary['min']) {
                        unset($values[$questionId]);
                    }

                    break;
                case 'number-between':
                    if (isset($values[$questionId]['from']) && $values[$questionId]['from'] == $boundary['min']
                        && isset($values[$questionId]['to']) && $values[$questionId]['to'] == $boundary['max']
                    ) {
                        unset($values[$questionId]);
                    }
            }
        }

        return apply_filters('wcsf_filtered_borderline_number_values', $values, $id);
    }
    
    /**
     * Get the filter step value list
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getValueList($id)
    {
        $value = self::getValue($id);
        $stepsIds = self::getStepsIds($id);
        $output = [];

        // sorting the values according the steps
        $value = array_replace(array_flip($stepsIds), $value);

        foreach ($value as $questionId => $valueItem) {
            if (!is_array($valueItem)) {
                continue;
            }

            $question = Question::get($questionId);

            if (!is_object($question) || !property_exists($question, 'viewType')) {
                continue;
            }

            $values = [];
            $anyProductValueText = Question::getSetting($questionId, 'any_product_value_text');
            $listItem = (object) [
                'id' => $questionId,
                'name' => wp_kses_post($question->name),
                'value' => null,
                'values' => []
            ];

            if ($question->viewType == 'formula') {
                foreach ($valueItem as $name => $item) {
                    $values[] = "$name: $item";
                }

                $listItem->value = implode(', ', $values);
                $listItem->values = $values;

                if (Question::getSetting($questionId, 'output_formula_value')) {
                    try {
                        $formulaValue = Core::executeFormula($question, $valueItem);
                        $listItem->value .= str_replace(
                            ['%value%'],
                            [$formulaValue],
                            Question::getSetting($questionId, 'formula_value_text')
                        );

                        $listItem->values[] = $formulaValue;
                    } catch (\Exception $exception) {
                        $listItem->value .= ' = ' . $exception->getMessage();
                        $listItem->values[] = $exception->getMessage();
                    }
                }
            } elseif (strpos($question->viewType, 'number') !== false) {
                $unit = Question::getSetting($questionId, 'numeric_unit');
                $fromText = Question::getSetting($questionId, 'from_value_text');
                $toText = Question::getSetting($questionId, 'to_value_text');

                // get numeric values
                if ($question->viewType == 'number' && !empty($valueItem)) {
                    $values[] = reset($valueItem) . $unit;
                }

                if (isset($valueItem['from']) && in_array($question->viewType, ['number-from', 'number-between'])) {
                    $values[] = "$fromText $valueItem[from]$unit";
                }

                if (isset($valueItem['to']) && in_array($question->viewType, ['number-to', 'number-between'])) {
                    $values[] = "$toText $valueItem[to]$unit";
                }

                $listItem->value = implode(' ', $values);
                $listItem->values = $values;
            } else {
                if ($question->filterType == 'manual') {
                    // get manual names
                    foreach ($valueItem as $valueItemPart) {
                        if (isset($question->manualValues[$valueItemPart])) {
                            $values[] = $question->manualValues[$valueItemPart]['name'];
                        }
                    }
                } elseif (in_array($question->filterType, ['tag', 'category', 'attribute'])) {
                    // define taxonomy
                    switch ($question->filterType) {
                        default:
                        case 'category':
                            $taxonomy = 'product_cat';
                            break;

                        case 'tag':
                            $taxonomy = 'product_tag';
                            break;

                        case 'attribute':
                            $taxonomy = \wc_attribute_taxonomy_name_by_id((int) $question->attribute);
                    }

                    // get terms values
                    foreach ($valueItem as $valueItemKey => $valueItemPart) {
                        // get multi-choice term name
                        if (strpos($question->viewType, 'multi-choice') !== false) {
                            if (in_array((int) $valueItemPart, [-1, 1, 2])) {
                                $term = get_term($valueItemKey, $taxonomy);

                                switch ((int) $valueItemPart) {
                                    case -1:
                                        $values[] = "<s class=\"wcsf-value-item-value-excluded\">$term->name</s>";
                                        break;

                                    case 1:
                                        $values[] = "<span class=\"wcsf-value-item-value-possible\">$term->name</span>";
                                        break;

                                    case 2:
                                    default:
                                        $values[] = "<b class=\"wcsf-value-item-value-required\">$term->name</b>";
                                }
                            }

                            continue;
                        }

                        // get "Doesn't matter" text
                        if ($valueItemPart == 0) {
                            $values[] = $anyProductValueText;

                            continue;
                        }

                        // get term name
                        $term = get_term($valueItemPart, $taxonomy);

                        if ($term) {
                            $values[] = $term->name;
                        }
                    }
                } elseif ($question->filterType == 'meta') {
                    foreach ($valueItem as $valueItemKey => $valueItemPart) {
                        // get multi-choice term name
                        if (strpos($question->viewType, 'multi-choice') !== false) {
                            if (in_array((int) $valueItemPart, [-1, 1, 2])) {
                                switch ((int) $valueItemPart) {
                                    case -1:
                                        $values[] = "<s class=\"wcsf-value-item-value-excluded\">$valueItemKey</s>";
                                        break;

                                    case 1:
                                        $values[] =
                                            "<span class=\"wcsf-value-item-value-possible\">$valueItemKey</span>";
                                        break;

                                    case 2:
                                    default:
                                        $values[] = "<b class=\"wcsf-value-item-value-required\">$valueItemKey</b>";
                                }
                            }

                            continue;
                        }

                        // get "Doesn't matter" text
                        if ($valueItemPart == '0') {
                            $values[] = $anyProductValueText;

                            continue;
                        }

                        $values[] = $valueItemPart;
                    }
                }

                $listItem->value = implode(', ', $values);
                $listItem->values = $values;
            }

            if (!empty($listItem->value)) {
                $output[] = $listItem;
            }
        }

        return apply_filters('wcsf_value_list', $output, $id);
    }
    // </editor-fold>

    // <editor-fold desc="Nav">
    /**
     * Get filter navigation items array
     *
     * @param integer $id
     *
     * @return array
     */
    public static function getNavItems($id)
    {
        $steps = self::getSteps($id);
        $activeStep = self::getActiveStepId($id);
        $nextActiveStepId = self::getNextActiveStepId($id);
        $mode = self::getSetting($id, 'mode');
        $navAction = self::getSetting($id, 'nav_action');
        $output = [];
        $metActiveItem = false;

        foreach ($steps as $step) {
            $isActiveStep = $activeStep == $step->id;
            $navItem = [
                'id' => $step->id,
                'name' => $step->name,
                'class' => '',
                'component' => "wcsf-$navAction",
                'type' => 'submit',
                'state' => 'default',
                'disabled' => false,
                'thumbnail' => $step->thumbnail
            ];

            if ($isActiveStep) {
                $metActiveItem = true;
            }

            if ($mode == 'step-by-step') {
                // step-by-step mode
                if ($isActiveStep) {
                    // is active step
                    $navItem['type'] = 'button';
                    $navItem['state'] = 'active';
                    $navItem['class'] = 'active';
                    $navItem['component'] = 'wcsf-none';
                } elseif ($nextActiveStepId == $step->id) {
                    // is next active step
                    $navItem['state'] = 'default';
                    $navItem['class'] = 'next-active';
                } else {
                    // the other steps
                    $navItem['state'] = $metActiveItem ? 'default' : 'disabled';
                    $navItem['class'] = $metActiveItem ? 'default' : 'past';
                    $navItem['disabled'] = $metActiveItem;
                    $navItem['component'] = $metActiveItem ? "wcsf-$navAction" : 'wcsf-get-step';
                }
            } else {
                // free walk mode
                $navItem['type'] = $isActiveStep ? 'button' : 'submit';
                $navItem['state'] = $isActiveStep ? 'active' : 'default';
                $navItem['class'] = $isActiveStep ? 'active' : ($metActiveItem ? 'default' : 'past');
                $navItem['component'] = $isActiveStep ? 'wcsf-none' : "wcsf-$navAction";
            }

            $output[] = $navItem;
        }

        return apply_filters('wcsf_nav_items', $output, $id);
    }

    /**
     * Get pagination items array for products loop
     *
     * @param array $args
     *
     * @return array
     */
    public static function getProductsPaginationItems($args)
    {
        $output = [];
        $defaults = [
            'page' => 1,
            'productsQuery' => []
        ];

        $args = array_replace_recursive($defaults, $args);

        if (!$args['productsQuery'] || empty($args['productsQuery'])) {
            return [];
        }

        $paginationArgs = apply_filters(
            'wcsf_products_pagination_args',
            [
                'format' => '?paged=%#%',
                'base' => '%_%',
                'total' => $args['productsQuery']->max_num_pages,
                'current' => max($args['page'], get_query_var('paged')),
                'show_all' => false,
                'end_size' => 1,
                'mid_size' => 2,
                'prev_next' => true,
                'prev_text'    => '&larr;',
                'next_text'    => '&rarr;',
                'type' => 'array'
            ],
            $args
        );

        $links = paginate_links($paginationArgs);

        foreach ((array) $links as $link) {
            // add custom classes
            $link = str_replace('page-numbers', 'page-numbers page-link', $link);

            // replace empty href
            $link = str_replace('href=""', 'href="?paged=1"', $link);
            $link = str_replace("href=''", 'href="?paged=1"', $link);

            preg_match_all('/<a[^>]+href=([\'"])(?<href>.+?)\1[^>]*>/i', $link, $result);

            if (!empty($result) && !empty($result['href'][0])) {
                $href = $result['href'][0];
                $linkParts = parse_url($href);

                parse_str($linkParts['query'], $linkPartsQuery);

                $paged = isset($linkPartsQuery['paged']) ? $linkPartsQuery['paged'] : 1;

                // add custom attributes
                $link = str_replace(
                    ' href=',
                    ' data-component="wcsf-products-pagination-link wcsf-get-step" data-step-id="results" data-page="'
                    . $paged . '" href=',
                    $link
                );
            }

            $output[] = (object) [
                'class' => strpos($link, 'current') !== false ? 'active' : '',
                'innerHtml' => $link
            ];
        }

        return apply_filters('wcsf_products_pagination_items', $output, $args);
    }

    /**
     * Get link URL for the shop "repeat" control
     *
     * @param integer $id
     * @param string $url
     *
     * @return string
     */
    public static function getShopRepeatUrl($id, $url)
    {
        $query = [];
        $parts = parse_url($url);

        // parse query args
        if (isset($parts['query'])) {
            parse_str($parts['query'], $query);
        }

        $query['wcsf-reset-state'] = $id;

        // build query string
        $parts['query'] = http_build_query($query);

        // stringify URL parts
        $output = Utils::buildUrl($parts);

        return apply_filters('wcsf_shop_repeat_url', $output, $url, $parts);
    }
    // </editor-fold>
    
    // <editor-fold desc="Results">
    /**
     * Get filter results URL with query args
     *
     * @param integer $id
     * @param array $queryArgs - extra query argrs
     *
     * @return string
     */
    public static function getResultsUrl($id, $queryArgs = [])
    {
        $url = self::getSetting($id, 'results_url');
        $query = [];

        // if the settings is empty
        if (!$url) {
            $url = get_permalink(wc_get_page_id('shop'));
        }

        // will stay on the same page
        if ($url == '#') {
            if (isset($queryArgs['wcsf-referer'])) {
                $url = $queryArgs['wcsf-referer'];
            } else {
                $url = (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on' ? 'https' : 'http')
                    . "://$_SERVER[HTTP_HOST]$_SERVER[REQUEST_URI]";
            }
        } else {
            $noResultsRedirectLink = self::getSetting($id, 'empty_results_url');
            $redirectToOnlyResultPage = self::getSetting($id, 'redirect_to_only_result_page');

            if ($noResultsRedirectLink || $redirectToOnlyResultPage) {
                $preliminaryResults = self::getPreliminaryResultsQuery($id);

                if ($redirectToOnlyResultPage && $preliminaryResults->post_count == 1) {
                    $url = get_permalink(reset($preliminaryResults->posts));
                } elseif ($noResultsRedirectLink && $preliminaryResults->post_count == 0) {
                    $url = $noResultsRedirectLink;
                }
            }
        }

        // if URL is absolute
        if (strpos($url, $_SERVER['HTTP_HOST']) === false) {
            $url = home_url() . '/' . $url;
        }

        $parts = parse_url($url);

        // parse query args
        if (isset($parts['query'])) {
            parse_str($parts['query'], $query);
        }

        // redirect to the product archive
        if (!empty($query['page_id']) && '' === get_option('permalink_structure')
            && wc_get_page_id('shop') == $query['page_id']
        ) {
            $url = get_post_type_archive_link('product');
            $parts = parse_url($url);
            $query = [];

            // parse query args
            if (isset($parts['query'])) {
                parse_str($parts['query'], $query);
            }
        }

        // add questions extra parameters
        foreach (self::getValue($id) as $questionId => $value) {
            $viewType = Question::getSetting($questionId, 'view_type');
            $filterType = Question::getSetting($questionId, 'filter_type');
            $additionalParam = Question::getSetting($questionId, 'additional_get_param');

            if (!$additionalParam || !trim($additionalParam) || strpos($viewType, 'multi-choice') !== false) {
                continue;
            }

            if (in_array($filterType, ['attribute', 'category', 'tag'])) {
                $additionalGetParamTermSource = Question::getSetting($questionId, 'additional_get_param_term_source');

                if (in_array($additionalGetParamTermSource, ['slug', 'name'])) {
                    switch ($filterType) {
                        default:
                        case 'category':
                            $taxonomy = 'product_cat';
                            break;

                        case 'tag':
                            $taxonomy = 'product_tag';
                            break;

                        case 'attribute':
                            $attributeId = Question::getSetting($questionId, 'attribute');
                            $attribute = wc_get_attribute($attributeId);
                            $taxonomy = $attribute->slug;
                    }

                    foreach ($value as &$valueItem) {
                        $term = get_term($valueItem, $taxonomy);

                        switch ($additionalGetParamTermSource) {
                            case 'slug':
                                $valueItem = $term->slug;
                                break;

                            case 'name':
                                $valueItem = $term->name;
                        }
                    }
                }
            }

            if (in_array($viewType, ['checkbox', 'image-checkbox', 'chips-checkbox', 'button-checkbox'])) {
                $query[$additionalParam] = $value;

                continue;
            }

            if ($viewType == 'number-between') {
                $additionalParam = explode('|', $additionalParam);

                if (count($additionalParam) > 1) {
                    foreach ($value as $valueItem) {
                        $query[array_shift($additionalParam)] = trim($valueItem);
                    }
                } else {
                    $query[array_shift($additionalParam)] = $value;
                }

                continue;
            }

            $query[$additionalParam] = reset($value);
        }

        // blend in required args
        $query['wcsf-results'] = $id;
        $query['wcsf-value'] = self::getValue($id);
        $query['wcsf-referer'] = (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on' ? 'https' : 'http')
            . "://$_SERVER[HTTP_HOST]$_SERVER[REQUEST_URI]";

        $query = array_replace($query, $queryArgs);

        // unset reset query
        if (isset($query['wcsf-reset-value'])) {
            unset($query['wcsf-reset-value']);
        }

        if (isset($query['wcsf-reset-state'])) {
            unset($query['wcsf-reset-state']);
        }

        // woof plugin query emulate
        if (class_exists('\\WOOF')) {
            $query['swoof'] = 1;
        }

        // apply outer filters
        $query = apply_filters('wcsf_results_redirect_url_query', $query, $id);

        // build query string
        $parts['query'] = http_build_query($query);

        return apply_filters('wcsf_results_redirect_url', Utils::buildUrl($parts), $id, $parts);
    }

    /**
     * Get filter reset URL
     *
     * @param integer $id
     * @param array $queryArgs - extra query argrs
     *
     * @return string
     */
    public static function getResetUrl($id, $queryArgs = [])
    {
        $query = [];

        if (isset($queryArgs['wcsf-referer'])) {
            $url = $queryArgs['wcsf-referer'];
        } else {
            $url = (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on' ? 'https' : 'http')
                . "://$_SERVER[HTTP_HOST]$_SERVER[REQUEST_URI]";
        }

        // if the settings is empty
        if (!$url) {
            $url = get_permalink(wc_get_page_id('shop'));
        }

        // if URL is absolute
        if (strpos($url, $_SERVER['HTTP_HOST']) === false) {
            $url = home_url() . '/' . $url;
        }

        $parts = parse_url($url);

        // parse query args
        if (isset($parts['query'])) {
            parse_str($parts['query'], $query);
        }

        // add reset state
        $query['wcsf-reset-state'] = $id;

        $query = array_replace($query, $queryArgs);

        // clear the query
        unset($query['wcsf-reset-value']);
        unset($query['wcsf-results']);
        unset($query['wcsf-value']);
        unset($query['wcsf-referer']);

        // apply outer filters
        $query = apply_filters('wcsf_results_reset_url_query', $query, $id);

        // build query string
        $parts['query'] = http_build_query($query);

        return apply_filters('wcsf_reset_redirect_url', Utils::buildUrl($parts), $id, $parts);
    }

    /**
     * Get preliminary results query by the current value
     *
     * @param integer $id
     * @param array $valuePart - extra value to merge with the whole current value
     * @param array $args - extra args for wp_query
     *
     * @return \WP_Query
     */
    public static function getPreliminaryResultsQuery($id, $valuePart = [], $args = [])
    {
        $defaultQueryArgs = [
            'post_type' => 'product',
            'fields' => 'ids',
            'no_found_rows' => true,
            'nopaging' => true
        ];

        $value = self::getValue($id);
        $value = array_replace_recursive($value, $valuePart);
        $queryArgs = Core::getQueryArgsByFilter($id, $value, $args);
        $queryArgs = array_replace_recursive($defaultQueryArgs, $queryArgs);
        $query = new \WP_Query($queryArgs);

        return $query;
    }

    /**
     * Get preliminary products IDs by the current value
     *
     * @param integer $id
     * @param array $valuePart - extra value to merge with the whole current value
     * @param array $args - extra args for wp_query
     *
     * @return array
     */
    public static function getPreliminaryResultsIds($id, $valuePart = [], $args = [])
    {
        $query = self::getPreliminaryResultsQuery($id, $valuePart, $args);

        return $query->posts;
    }

    /**
     * Get preliminary products count by the current value
     *
     * @param integer $id
     * @param array $valuePart - extra value to merge with the whole current value
     * @param array $args - extra args for wp_query
     *
     * @return integer
     */
    public static function getPreliminaryResultsCount($id, $valuePart = [], $args = [])
    {
        $query = self::getPreliminaryResultsQuery($id, $valuePart, $args);

        return (int) $query->post_count;
    }

    /**
     * Get products request arguments
     *
     * @param integer $id
     * @param array $args
     *
     * @return array
     */
    public static function getProductsQueryArgs($id, $args = [])
    {
        $defaults = [
            'orderby' => 'menu_order',
            'page' => 1
        ];

        $args = array_replace($defaults, $args);
        $orderParts = explode('-', $args['orderby']);
        $order = isset($orderParts[1]) ? $orderParts[1] : 'asc';
        $value = self::getValue($id);
        $output = Core::getQueryArgsByFilter($id, $value, $args);
        $wcQuery = new \WC_Query();
        $orderingArgs = $wcQuery->get_catalog_ordering_args($orderParts[0], $order);
        $defaults = [
            'post_type' => 'product',
            'posts_per_page' => wc_get_default_products_per_row() * wc_get_default_product_rows_per_page(),
            'orderby' => $args['orderby'],
            'paged' => $args['page'],
            'post_status' => 'publish',
            'tax_query' => []
        ];

        if (filter_var(get_option('woocommerce_hide_out_of_stock_items'), FILTER_VALIDATE_BOOLEAN)) {
            $defaults['tax_query'][] = [
                'taxonomy' => 'product_visibility',
                'field' => 'name',
                'terms' => 'outofstock',
                'operator' => 'NOT IN'
            ];
        }

        $output = array_replace_recursive($defaults, $args, $output, $orderingArgs);

        return apply_filters('wcsf_products_query_args', $output, $id, $args);
    }
    // </editor-fold>

    // <editor-fold desc="Results PDF">
    /**
     * Generate and return URL and path for cart content PDF
     *
     * @param integer $id
     *
     * @return null|object
     */
    public function getResultPDFInstance($id)
    {
        require_once __DIR__ . '/../vendor/dompdf/autoload.inc.php';

        do_action('wcsf_before_output', ['id' => $id]);

        $options = new \Dompdf\Options();
        do_action('wcsf_dompdf_options', $options, $id);

        $dompdf = new \Dompdf\Dompdf($options);
        do_action('wcsf_dompdf_instance', $dompdf, $id);

        $dompdf->loadHtml(Core::getTemplatePart('products/pdf', ['id' => $id], ['echo' => false]));
        $dompdf->render();
        $pageTotal = (int) $dompdf->getCanvas()->get_page_count();
        Instance()->currentResultsPDFTotalPages = $pageTotal;

        $dompdf = new \Dompdf\Dompdf();
        $dompdf->loadHtml(
            Core::getTemplatePart(
                'products/pdf',
                [
                    'id' => $id,
                    'pageTotal' => $pageTotal
                ],
                ['echo' => false]
            )
        );

        $dompdf->render();
        do_action('wcsf_after_output', ['id' => $id]);

        return apply_filters('wcsf_results_pdf_instance', $dompdf, $id);
    }

    /**
     * Save Results PDF to a file and return its URL and path
     *
     * @param integer $id
     *
     * @return null|array
     */
    public function saveAndGetResultPDF($id)
    {
        $dompdf = self::getResultPDFInstance($id);

        if (!$dompdf) {
            return null;
        }

        $name = apply_filters('wcsf_results_pdf_file_name', get_bloginfo('name'));
        $path = WC_STEP_FILTER_UPLOADS_PATH . DIRECTORY_SEPARATOR . 'pdf' . DIRECTORY_SEPARATOR . "$name.pdf";
        $url = WC_STEP_FILTER_UPLOADS_URL . DIRECTORY_SEPARATOR . 'pdf' . DIRECTORY_SEPARATOR . "$name.pdf";

        // create uploads folder if not exists
        if (!file_exists(WC_STEP_FILTER_UPLOADS_PATH)) {
            mkdir(WC_STEP_FILTER_UPLOADS_PATH, 0777, true);
        }

        // create pdf folder if not exists
        if (!file_exists(WC_STEP_FILTER_UPLOADS_PATH . DIRECTORY_SEPARATOR . 'pdf')) {
            mkdir(WC_STEP_FILTER_UPLOADS_PATH . DIRECTORY_SEPARATOR . 'pdf', 0777, true);
        }

        file_put_contents($path, $dompdf->output());

        $data = [
            'url' => $url,
            'path' => $path,
            'pdf' => $dompdf
        ];

        return apply_filters('wcsf_results_pdf_file', $data, $id);
    }

    /**
     * Output results PDF
     *
     * @param integer $id
     */
    public function outputResultPDF($id)
    {
        $dompdf = $this->getResultPDFInstance($id);

        if (!$dompdf) {
            return;
        }

        $name = apply_filters('wcsf_results_pdf_file_name', get_bloginfo('name'));
        $dompdf->stream($name, ['Attachment' => false]);

        exit();
    }
    // </editor-fold>

    // <editor-fold desc="Get">
    /**
     * Returns array of ids and titles of posts
     *
     * @param array $args
     *
     * @return array
     */
    public static function getPostsIdsList($args = [])
    {
        $defaults = [
            'post_type' => 'wcsf',
            'numberposts' => -1,
            'post_status' => 'publish'
        ];

        $args = array_replace($defaults, $args);
        $output = [];
        $posts = get_posts($args);

        foreach ($posts as $post) {
            $output[$post->ID] = $post->post_title;
        }

        return $output;
    }
    // </editor-fold>

    // <editor-fold desc="AJAX">
    /** Save value and get form template by ajax */
    public function submitAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null,
            'page' => 1,
            'referer' => home_url()
        ];

        $args = array_replace($defaults, $postData);
        $args['value'] = isset($args['value']) ? (array) $args['value'] : [$args['stepId'] => []];
        $args['resetUrl'] = self::getResetUrl($args['id'], ['wcsf-referer' => $args['referer']]);

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            Utils::sendJSON([
                'error' => true,
                'message' => $exception->getMessage(),
                'content' => Core::getTemplatePart('form', $args, ['echo' => false])
            ]);
        }

        Utils::sendJSON(['content' => Core::getTemplatePart('form', $args, ['echo' => false])]);
    }

    /** Save value and get a step form template by ajax */
    public function submitAndGoToAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null,
            'referer' => home_url(),
            'value' => []
        ];

        $args = array_replace($defaults, $postData);
        $args['resetUrl'] = self::getResetUrl($args['id'], ['wcsf-referer' => $args['referer']]);

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            Utils::sendJSON([
                'error' => true,
                'message' => $exception->getMessage(),
                'content' => Core::getTemplatePart('form', $args, ['echo' => false])
            ]);
        }

        // change step
        if ($args['stepId']) {
            self::setActiveStep($args['id'], $args['stepId']);
        }

        Utils::sendJSON(['content' => Core::getTemplatePart('form', $args, ['echo' => false])]);
    }

    /** Save value and get the next form template by ajax */
    public function submitAndGoNextAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null,
            'referer' => home_url(),
            'value' => []
        ];

        $args = array_replace($defaults, $postData);
        $args['resetUrl'] = self::getResetUrl($args['id'], ['wcsf-referer' => $args['referer']]);
        $showResultsWithin = self::getSetting($args['id'], 'show_results_within');

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            Utils::sendJSON([
                'error' => true,
                'message' => $exception->getMessage(),
                'content' => Core::getTemplatePart('form', $args, ['echo' => false])
            ]);
        }

        // get steps after value update
        $stepsIds = self::getStepsIds($args['id']);

        if (!$showResultsWithin && end($stepsIds) == $args['stepId']) {
            Utils::sendJSON(['redirect' => self::getResultsUrl($args['id'], ['wcsf-referer' => $args['referer']])]);
        }

        if (end($stepsIds) != $args['stepId']) {
            // set active step from the step id and increment it
            self::setActiveStep($args['id'], $args['stepId']);
            self::setActiveStep($args['id'], self::getNextActiveStepId($args['id']));
        }

        Utils::sendJSON(['content' => Core::getTemplatePart('form', $args, ['echo' => false])]);
    }

    /** Save value and get the results by ajax */
    public function submitAndGoResultsAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'referer' => home_url(),
            'value' => []
        ];

        $args = array_replace($defaults, $postData);
        $stepsIds = self::getStepsIds($args['id']);
        $showResultsWithin = self::getSetting($args['id'], 'show_results_within');
        $args['resetUrl'] = self::getResetUrl($args['id'], ['wcsf-referer' => $args['referer']]);

        try {
            // save value
            self::setValue($args['id'], $args['value']);
        } catch (\Exception $exception) {
            $this->addNotice(
                $exception->getCode() ? $exception->getCode() : self::getActiveStepId($postData['id']),
                ['message' => $exception->getMessage()]
            );

            Utils::sendJSON([
                'error' => true,
                'message' => $exception->getMessage(),
                'content' => Core::getTemplatePart('form', $args, ['echo' => false])
            ]);
        }

        // set last step as active
        self::setActiveStep($args['id'], end($stepsIds));

        if (!$showResultsWithin) {
            Utils::sendJSON(['redirect' => self::getResultsUrl($args['id'], ['wcsf-referer' => $args['referer']])]);
        }

        Utils::sendJSON(['content' => Core::getTemplatePart('form', $args, ['echo' => false])]);
    }

    /** Save value and get the form template by ajax */
    public function getStepHtmlAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'stepId' => null,
            'referer' => home_url(),
        ];

        $args = array_replace($defaults, $postData);
        $args['resetUrl'] = self::getResetUrl($args['id'], ['wcsf-referer' => $args['referer']]);

        // change step
        if ($args['stepId']) {
            self::setActiveStep($args['id'], $args['stepId']);
        }

        Utils::sendJSON(['content' => Core::getTemplatePart('form', $args, ['echo' => false])]);
    }

    /** Reset filter by ajax */
    public function resetAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'resetStepId' => null
        ];

        $args = array_replace($defaults, $postData);

        if (!$args['id']) {
            return;
        }

        self::resetState($args['id'], $args['resetStepId']);

        Utils::sendJSON(['content' => Core::getTemplatePart('form', ['id' => $args['id']], ['echo' => false])]);
    }

    /** Get preliminary results count by ajax */
    public function getPreliminaryResultsCountAjax()
    {
        $postData = Utils::parseArrayOfJSONs($_POST);
        $defaults = [
            'id' => null,
            'value' => [],
            'categoryId' => null,
            'tagId' => null
        ];

        $args = array_replace($defaults, $postData);
        $query = [];

        if ($args['categoryId']) {
            $query['categoryId'] = $args['categoryId'];
        }

        if ($args['tagId']) {
            $query['tagId'] = $args['tagId'];
        }

        $preliminaryResultsCountText = self::getSetting($args['id'], 'preliminary_results_count_text');
        $count = self::getPreliminaryResultsCount($args['id'], (array) $args['value'], $query);

        Utils::sendJSON([
            'message' => sprintf($preliminaryResultsCountText, $count),
            'value' => $count
        ]);
    }
    // </editor-fold>
}
