<?php
namespace WCStepFilter;

/**
 * Inquiry Class
 *
 * @class Inquiry
 * @version 1.3.0
 */
class Inquiry
{
    // <editor-fold desc="Properties">
    /**
     * Multi-choice question values
     *
     * @var array
     */
    public static $multiChoicePossibleValues;
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Constructor */
    public function __construct()
    {
        add_action('init', [$this, 'onInit'], 5);
        add_action('wcsf_before_results_redirect', [$this, 'beforeResultsRedirectAction']);
        add_action('wcsf_products_query', [$this, 'productsQueryAction'], 10, 3);

        // ajax
        add_action('wp_ajax_WCSFGetInquiriesListTable', [$this, 'outputListTableAjax']);

        // admin
        add_action('admin_menu', [$this, 'registerMenuItem']);
        add_filter('pre_get_posts', [$this, 'preGetPostsFilter']);
    }

    /** On init action */
    public function onInit()
    {
        self::$multiChoicePossibleValues = apply_filters(
            'wsf_multi_choice_possible_values',
            [
                0 => L10N::r('Undefined'),
                1 => L10N::r('Possible'),
                2 => L10N::r('Required'),
                -1 => L10N::r('Excluded')
            ]
        );

        $name = 'SF Inquiries';
        $args = [
            'label' => $name,
            'labels' => [
                'name' => $name,
                'singular_name' => $name,
                'menu_name' => $name
            ],
            'public' => false,
            'show_ui' => true,
            'map_meta_cap' => true,
            'publicly_queryable' => false,
            'exclude_from_search' => true,
            'show_in_menu' => false,
            'hierarchical' => false,
            'rewrite' => false,
            'query_var' => false,
            'supports' => [],
            'show_in_nav_menus' => false,
            'show_in_admin_bar' => false
        ];

        register_post_type('wcsf_inquiry', $args);

        // purge posts
        if (isset($_POST['woocommerce-step-filter'])
            && wp_verify_nonce($_POST['woocommerce-step-filter'], basename(__FILE__))
        ) {
            if (isset($_POST['purge-filter-inquiries']) && !empty($_POST['purge-filter-inquiries'])) {
                $filterId = (int) $_POST['purge-filter-inquiries'];
                $list = self::getList(['post_parent' => $filterId]);

                foreach ($list as $item) {
                    wp_delete_post($item->id);
                }
            }
        }
    }

    /**
     * Fires before redirect to the results page
     *
     * @param integer $filterId
     */
    public function beforeResultsRedirectAction($filterId)
    {
        if (empty($filterId) || !Filter::getSetting($filterId, 'collect_inquiries')) {
            return;
        }

        $value = Filter::getValue($filterId);

        if (empty($value)) {
            return;
        }

        self::add($filterId, $value, Filter::getPreliminaryResultsIds($filterId));
    }

    /**
     * Products within filter query
     *
     * @param integer $filterId
     * @param array $arguments
     * @param \WP_Query $productsQuery
     *
     * @return int|\WP_Error
     */
    public function productsQueryAction($filterId, $arguments, $productsQuery)
    {
        if (empty($filterId) || !Filter::getSetting($filterId, 'collect_inquiries')) {
            return 0;
        }

        $value = Filter::getValue($filterId);

        if (empty($value)) {
            return 0;
        }

        return self::add($filterId, $value, wp_list_pluck($productsQuery->posts, 'ID'));
    }

    /**
     * Add an inquiry record
     *
     * @param integer $filterId
     * @param array $value
     * @param array $productsIDs
     *
     * @return int|\WP_Error
     */
    public static function add($filterId, $value, $productsIDs = [])
    {
        $data = [
            'post_author' => 1, // bind all records to the first user
            'post_type' => 'wcsf_inquiry',
            'post_parent' => (int) $filterId,
            'post_content' => serialize($value),
            'post_excerpt' => serialize($productsIDs),
            'menu_order' => count($productsIDs)
        ];

        $data = apply_filters('wcsf_add_inquiry_data', $data, $filterId, $value, $productsIDs);

        do_action('wcsf_before_add_inquiry', $data, $filterId, $value, $productsIDs);

        return wp_insert_post($data);
    }

    /**
     * Update inquiry record
     *
     * @param integer $id
     * @param array $args
     *
     * @return int|\WP_Error
     */
    public static function update($id, $args)
    {
        $defaults = [
            'filterId' => null,
            'value' => null,
            'resultsIds' => null,
            'resultsCount' => null
        ];

        $args = array_replace($defaults, $args);
        $data = ['ID' => $id];

        if (!is_null($args['filterId'])) {
            $data['post_parent'] = (int) $args['filterId'];
        }

        if (!is_null($args['value'])) {
            $data['post_content'] = serialize($args['value']);
        }

        if (!is_null($args['resultsIds'])) {
            $data['post_excerpt'] = serialize($args['resultsIds']);
        }

        if (!is_null($args['resultsCount'])) {
            $data['menu_order'] = (int) $args['resultsCount'];
        }

        $data = apply_filters('wcsf_update_inquiry_data', $data, $id, $args);

        return wp_update_post($data);
    }
    // </editor-fold>

    // <editor-fold desc="Admin">
    /** Register settings page menu item action */
    public function registerMenuItem()
    {
        add_submenu_page(
            'edit.php?post_type=wcsf',
            L10N::r('Inquiries'),
            L10N::r('Inquiries'),
            'manage_options',
            'woocommerce-step-filter-inquiries',
            [$this, 'adminPageCallback']
        );
    }

    /**
     * Filter pre get post query
     *
     * @param \WP_Query $query
     *
     * @return \WP_Query
     */
    public function preGetPostsFilter($query)
    {
        if (!is_admin()) {
            return $query;
        }

        global $pagenow;

        if ('edit.php' !== $pagenow) {
            return $query;
        }

        if (!isset($_GET['wcsf_ids']) || empty($_GET['wcsf_ids'])) {
            return $query;
        }

        $query->set('post__in', wp_parse_id_list($_GET['wcsf_ids']));

        return $query;
    }

    /** Plugins global settings page view */
    public function adminPageCallback()
    {
        $filters = Filter::getPostsIdsList();
        $filterId = isset($_GET['filter-id'])
            ? (int) $_GET['filter-id']
            : (isset(array_keys($filters)[0]) ? array_keys($filters)[0] : null);

        $statistic = $filterId ? self::getStatisticData($filterId) : null;
        ?>
        <div class="wrap" data-component="wcsf-inquire">
            <h1><?php echo esc_html(get_admin_page_title()); ?></h1>
            <form action="?page=woocommerce-step-filter-inquiries" method="get">
                <input type="hidden" name="post_type" value="wcsf">
                <input type="hidden" name="page"
                    value="<?php echo isset($_GET['page']) ? $_GET['page'] : 'woocommerce-step-filter-inquiries'; ?>">
                <table class="form-table">
                    <tbody>
                        <tr>
                            <th><label for="wcsf-filter-id"><?php L10N::e('Filter'); ?></label></th>
                            <td>
                                <select name="filter-id"
                                    id="wcsf-filter-id"><?php
                                    foreach ($filters as $id => $name) {
                                        echo '<option value="' . esc_attr($id) . '" '
                                            . ($filterId == $id ? ' selected' : '') . '>'
                                            . esc_html($name)
                                            . '</option>';
                                    }
                                    ?></select>
                                <button class="button button-primary"
                                    type="submit"><?php L10N::e('Get'); ?></button>
                            </td>
                        </tr>
                    </tbody>
                </table>
            </form>
            <?php
            $this->outputQuestions($statistic);

            if ($statistic) {
                $this->outputList($filterId, $statistic->totalInquiries);
            }
            ?>
        </div>
        <?php
    }

    /**
     * Output list HTML
     *
     * @param integer $filterId
     * @param integer $total
     */
    public function outputList($filterId, $total)
    {
        $inquiries = self::getList([
            'post_parent' => $filterId,
            'numberposts' => 1
        ]);
        
        if (empty($inquiries)) {
            return;
        }
        ?>
        <section class="wcsf-inquiry-list" data-component="wcsf-inquiry-list">
            <h2 class="wcsf-inquiry-list-title"><?php L10N::e('Inquires list'); ?></h2>
            <?php $this->outputListTable($filterId, $total); ?>
            <form action="#" method="post">
                <?php wp_nonce_field(basename(__FILE__), 'woocommerce-step-filter'); ?>
                <table class="wp-list-table widefat">
                    <tr>
                        <td>
                            <button class="button"
                                data-component="wcsf-inquiries-purge"
                                name="purge-filter-inquiries"
                                value="<?php echo esc_attr($filterId); ?>"><?php
                                L10N::e('Purge all data');
                                ?></button>
                        </td>
                    </tr>
                </table>
            </form>
        </section>
        <?php
    }

    /**
     * Output list table HTML
     *
     * @param integer $filterId
     * @param integer $total
     * @param integer $paged
     */
    public function outputListTable($filterId, $total, $paged = null)
    {
        if (is_null($paged)) {
            $paged = isset($_GET['paged']) ? $_GET['paged'] : 1;
        }

        $perPage = apply_filters('wcsf_inquiries_list_per_page', 20, $filterId, $total, $paged);
        $questions = Filter::getQuestionsList($filterId);
        $inquiries = self::getList([
            'post_parent' => $filterId,
            'numberposts' => $perPage,
            'offset' => ($paged - 1) * $perPage
        ]);
        
        $pagination = self::getListPagination([
            'page' => $paged,
            'total' => ceil($total / $perPage)
        ]);
        ?>
        <table class="wcsf-inquiry-list-table wp-list-table widefat striped"
            data-component="wcsf-inquiry-list-table"
            data-filter-id="<?php echo esc_attr($filterId); ?>"
            data-total="<?php echo esc_attr($total); ?>"
            data-ajax-url="<?php echo admin_url('admin-ajax.php'); ?>">
            <thead>
                <tr>
                    <th><?php L10N::e('ID'); ?></th>
                    <th><?php L10N::e('Date'); ?></th>
                    <th><?php L10N::e('Results count'); ?></th>
                    <th><?php L10N::e('Values'); ?></th>
                </tr>
            </thead>
            <tbody>
                <?php foreach ($inquiries as $inquiry) { ?>
                    <tr class="wcsf-inquiry-list-item">
                        <td><?php echo wp_kses_post($inquiry->id); ?></td>
                        <td><?php echo wp_kses_post($inquiry->date); ?></td>
                        <td>
                            <a target="_blank"
                                href="<?php
                                echo get_admin_url(
                                    null,
                                    'edit.php?post_type=product&wcsf_ids=' . implode(',', $inquiry->resultsIds)
                                );
                                ?>"><?php echo wp_kses_post($inquiry->resultsCount); ?></a>
                        </td>
                        <td>
                            <dl class="wcsf-inquiry-list-item-data">
                                <?php
                                foreach ($inquiry->values as $questionId => $values) {
                                    if (!isset($questions[$questionId])) {
                                        continue;
                                    }

                                    $output = [];
                                    $fields = Question::getFields($filterId, $questionId);
                                    $question = $questions[$questionId];
                                    $name = $question->name;
                                    $isNumericView = strpos($question->viewType, 'number') !== false;
                                    $isMultiChoiceView
                                        = strpos($question->viewType, 'multi-choice') !== false;

                                    echo '<div class="wcsf-inquiry-list-item-data-item">';
                                    echo '<dt class="wcsf-inquiry-list-item-data-item-name">'
                                        . wp_kses_post($name)
                                        . ':</dt> ';

                                    foreach ($values as $key => $value) {
                                        foreach ($fields as $fieldKey => $field) {
                                            if ((!$isNumericView && !$isMultiChoiceView && $field->value != $value)
                                                || ($isNumericView && $key != $fieldKey)
                                                || ($isMultiChoiceView && $field->value != $key)
                                            ) {
                                                continue;
                                            }

                                            $valueOutput = $field->name;

                                            if ($isNumericView) {
                                                $valueOutput = $value;
                                            } elseif ($isMultiChoiceView) {
                                                $valueOutput =
                                                    "<b>$field->name"
                                                    . '</b>: <span>'
                                                    . self::$multiChoicePossibleValues[$value]
                                                    . '</span>';
                                            }

                                            $output[] = '<dd class="wcsf-inquiry-list-item-data-item-value">'
                                                . wp_kses_post($valueOutput)
                                                . '</dd> ';
                                        }
                                    }

                                    if ($isNumericView) {
                                        echo implode(' - ', $output);
                                    } else {
                                        echo implode(', ', $output);
                                    }

                                    echo '</div>';
                                }
                                ?>
                            </dl>
                        </td>
                    </tr>
                <?php } ?>
            </tbody>
            <?php if (!empty($pagination)) { ?>
                <tfoot>
                    <tr>
                        <td colspan="4">
                            <nav class="wcsf-inquiry-list-pagination">
                                <ul class="wcsf-inquiry-list-pagination-list pagination"
                                    data-component="wcsf-inquiry-list-pagination">
                                    <?php foreach ($pagination as $item) { ?>
                                        <li class="<?php
                                        echo esc_attr(trim("wcsf-inquiry-list-pagination-list-item $item[class]"));
                                        ?>"><?php echo $item['innerHtml']; ?></li>
                                    <?php } ?>
                                </ul>
                            </nav>
                        </td>
                    </tr>
                </tfoot>
            <?php } ?>
        </table>
        <?php
    }

    /** Output list table HTML via Ajax */
    public function outputListTableAjax()
    {
        $filterId = isset($_GET['filterId']) ? (int) $_GET['filterId'] : null;
        $total = isset($_GET['total']) ? (int) $_GET['total'] : null;
        $page = isset($_GET['page']) ? (int) $_GET['page'] : null;

        $this->outputListTable($filterId, $total, $page);

        exit;
    }

    /**
     * Get pagination items array
     *
     * @param array $args
     *
     * @return array
     */
    public static function getListPagination($args)
    {
        $defaults = [
            'page' => 1,
            'total' => 10
        ];

        $args = array_merge($defaults, $args);
        $output = [];
        $paginationArgs = apply_filters(
            'wcsf_inquiries_list_pagination_args',
            [
                'format' => '?paged=%#%',
                'base' => '%_%',
                'total' => $args['total'],
                'current' => $args['page'],
                'show_all' => false,
                'end_size' => 1,
                'mid_size' => 2,
                'prev_next' => true,
                'prev_text' => L10N::r('« Previous'),
                'next_text' => L10N::r('Next »'),
                'type' => 'array'
            ],
            $args
        );

        $links = paginate_links($paginationArgs);

        foreach ((array) $links as $link) {
            // add custom classes
            $link = str_replace(['page-numbers', 'current'], ['page-numbers button', 'current active'], $link);

            // replace empty href
            $link = str_replace('href=""', 'href="?paged=1"', $link);
            $link = str_replace("href=''", 'href="?paged=1"', $link);

            preg_match_all('/<a[^>]+href=([\'"])(?<href>.+?)\1[^>]*>/i', $link, $result);

            if (!empty($result) && !empty($result['href'][0])) {
                $href = $result['href'][0];
                $linkParts = parse_url($href);

                parse_str($linkParts['query'], $linkPartsQuery);

                $paged = isset($linkPartsQuery['paged']) ? $linkPartsQuery['paged'] : 1;

                // add custom attributes
                $link = str_replace(
                    ' href=',
                    ' data-component="wcsf-inquiry-list-pagination-item-link" data-page="' . $paged . '" href=',
                    $link
                );
            }

            $output[] = [
                'class' => strpos($link, 'current') !== false ? 'active' : '',
                'innerHtml' => $link
            ];
        }

        return apply_filters('wcsf_inquiries_list_pagination', $output, $args);
    }

    /**
     * Output questions statistic HTML
     *
     * @param object $statistic - @see getStatisticData
     */
    public function outputQuestions($statistic)
    {
        if (is_null($statistic)) {
            return;
        }
        ?>
        <section class="wcsf-inquiry-questions">
            <h2 class="wcsf-inquiry-questions-title"><?php
                printf(L10N::r('Inquires count: %d'), $statistic->totalInquiries);
                ?></h2>
            <?php foreach ($statistic->questions as $question) { ?>
                <section class="wcsf-inquiry-question">
                    <h3 class="wcsf-inquiry-question-title"><?php
                        echo wp_kses_post($question->name);
                        ?></h3>
                    <div class="wcsf-inquiry-question-fields-wrapper">
                        <div class="wcsf-inquiry-question-fields">
                            <?php foreach ($question->values as $value) { ?>
                                <div class="wcsf-inquiry-question-fields-item">
                                    <div class="wcsf-inquiry-question-fields-item-inner"
                                        style="<?php echo esc_attr("height:$value->percent%;"); ?>">
                                        <div class="wcsf-inquiry-question-fields-item-content"
                                            title="<?php echo esc_attr("$value->name ($value->count)"); ?>">
                                            <span class="wcsf-inquiry-question-fields-item-name"><?php
                                                echo wp_kses_post($value->name);
                                                ?></span>
                                            /
                                            <span class="wcsf-inquiry-question-fields-item-percent"><?php
                                                echo wp_kses_post(round($value->percent, 2));
                                                ?>%</span>
                                            <small class="wcsf-inquiry-question-fields-item-count"><?php
                                                echo wp_kses_post("($value->count)");
                                                ?></small>
                                        </div>
                                    </div>
                                </div>
                            <?php } ?>
                        </div>
                    </div>
                </section>
            <?php } ?>
        </section>
        <?php
    }
    // </editor-fold>

    // <editor-fold desc="Get">
    /**
     * Return inquiry data object
     *
     * @param object $inquiry - WP query post
     *
     * @return object
     */
    public static function getData($inquiry)
    {
        $output = (object) [
            'id' => (int) $inquiry->ID,
            'date' => $inquiry->post_date,
            'filterId' => (int) $inquiry->post_parent,
            'resultsCount' => (int) $inquiry->menu_order,
            'resultsIds' => (array) unserialize($inquiry->post_excerpt),
            'values' => (array) unserialize($inquiry->post_content)
        ];

        return apply_filters('wcsf_inquiry_data', $output, $inquiry);
    }

    /**
     * Return array of inquires
     *
     * @param array $args
     *
     * @return array
     */
    public static function getList($args)
    {
        $defaults = [
            'post_type' => 'wcsf_inquiry',
            'post_status' => 'any',
            'numberposts' => 30
        ];

        $args = array_replace($defaults, $args);
        $inquires = get_posts($args);
        $output = [];

        foreach ($inquires as $inquiry) {
            $output[$inquiry->ID] = self::getData($inquiry);
        }

        return apply_filters('wcsf_inquires_list', $output, $args);
    }

    /**
     * Returns array of ids and titles of posts
     *
     * @param array $args
     *
     * @return array
     */
    public static function getPostsIdsList($args = [])
    {
        $defaults = [
            'post_type' => 'wcsf_inquiry',
            'numberposts' => -1,
            'post_status' => 'publish'
        ];

        $args = array_replace($defaults, $args);
        $output = [];
        $posts = get_posts($args);

        foreach ($posts as $post) {
            $output[$post->ID] = $post->post_title;
        }

        return $output;
    }

    /**
     * Returns object of the filter's statistic data
     *
     * @param integer $filterId
     *
     * @return object
     */
    public static function getStatisticData($filterId)
    {
        $output = [];
        $questions = Filter::getQuestionsList($filterId);
        $inquiriesAll = self::getList([
            'post_parent' => $filterId,
            'numberposts' => -1
        ]);

        foreach ($questions as $question) {
            $fields = Question::getFields($filterId, $question->id);
            $isNumericView = strpos($question->viewType, 'number') !== false;
            $isMultiChoiceView = strpos($question->viewType, 'multi-choice') !== false;
            $outputItems = [];

            if ($isNumericView) {
                foreach ($fields as $key => $field) {
                    $outputItems[$question->id . '-' . $key] = (object) [
                        'name' => $question->name . ': ' . $field->label,
                        'values' => []
                    ];
                }
            } elseif ($isMultiChoiceView) {
                foreach ($fields as $field) {
                    $outputItems[$question->id . '-' . $field->name] = (object) [
                        'name' => $question->name . ': ' . $field->name,
                        'values' => []
                    ];
                }
            } else {
                $outputItems[$question->id] = (object) [
                    'name' => $question->name,
                    'values' => []
                ];

                // fill by all exist fields
                foreach ($fields as $field) {
                    $outputItems[$question->id]->values[$field->value] = 0;
                }
            }

            foreach ($fields as $field) {
                foreach ($inquiriesAll as $inquiry) {
                    // have no this question in the inquiry answer
                    if (!isset($inquiry->values[$question->id])) {
                        continue;
                    }

                    $inquiryValues = $inquiry->values[$question->id];

                    if ($isNumericView) {
                        if (is_array($inquiryValues)) {
                            foreach ($inquiryValues as $key => $inquiryValue) {
                                if (!isset($outputItems[$question->id . '-' . $key]->values[$inquiryValue])) {
                                    $outputItems[$question->id . '-' . $key]->values[$inquiryValue] = 0;
                                }

                                $outputItems[$question->id . '-' . $key]->values[$inquiryValue]++;
                            }
                        }
                    } elseif ($isMultiChoiceView) {
                        if (!isset($inquiryValues[$field->value])) {
                            continue;
                        }

                        foreach (self::$multiChoicePossibleValues as $key => $label) {
                            if (!isset($outputItems[$question->id . '-' . $field->name]->values[$label])) {
                                $outputItems[$question->id . '-' . $field->name]->values[$label] = 0;
                            }
                        }

                        $label = self::$multiChoicePossibleValues[$inquiryValues[$field->value]];
                        $outputItems[$question->id . '-' . $field->name]->values[$label]++;
                    } elseif (in_array($field->value, $inquiryValues)) {
                        $outputItems[$question->id]->values[$field->value]++;
                    }
                }
            }

            foreach ($outputItems as $key => &$outputItem) {
                if (!property_exists($outputItem, 'name')) {
                    unset($outputItems[$key]);

                    continue;
                }

                // add other properties
                $outputItem->fields = $fields;
                $outputItem->totalValuesCount = array_sum($outputItem->values);

                foreach ($outputItem->values as $value => $count) {
                    $name = $value;
                    $percent = $outputItem->totalValuesCount != 0
                        ? ($count / $outputItem->totalValuesCount) * 100
                        : 0;

                    foreach ($outputItem->fields as $field) {
                        if ($field->value === $value) {
                            $name = $field->name;

                            break;
                        }
                    }

                    $outputItem->values[$value] = (object) [
                        'count' => $count,
                        'name' => $name,
                        'percent' => $percent
                    ];
                }

                // sort numeric questions
                if ($isNumericView) {
                    ksort($outputItem->values);
                }

                unset($outputItem);
            }

            $output = array_merge($output, $outputItems);
        }

        return (object) [
            'questions' => $output,
            'totalInquiries' => count($inquiriesAll)
        ];
    }
    // </editor-fold>
}
