/* Woocommerce Step Filter
 * Original author: troll_winner@mail.ru
 * Further changes, comments: troll_winner@mail.ru
 */
(function (root, factory) {
    'use strict';

    if (typeof define === 'function' && define.amd) {
        define(['jquery', 'jquery.serializejson'], factory);
    } else if (typeof exports === 'object'
        && typeof module !== 'undefined'
        && typeof require === 'function'
    ) {
        module.exports = factory(require('jquery'), require('jquery.serializejson'));
    } else {
        factory(root.jQuery);
    }
})(this, function ($) {
    'use strict';

    const pluginName = 'wcsf';
    const defaults = {
        documentNode: document,
        windowNode: window,
        bodySelector: 'html, body',
        preliminaryResultsCountRequestInterval: 1000
    };

    const Plugin = function (element, options) {
        this.element = element;
        this.options = $.extend({}, defaults, options);

        return this.init();
    };

    /**
     * Init the instance
     * @returns {Plugin} self instance
     */
    Plugin.prototype.init = function () {
        this.$element = $(this.element);
        this.$document = $(this.options.documentNode);
        this.$window = $(this.options.windowNode);
        this.$body = $(this.options.bodySelector);
        this.args = this.$element.data('arguments');
        this.getPreliminaryResultsCountTimeout = null;
        this.getPreliminaryResultsCountPromise = null;

        return this;
    };

    /**
     * Makes an ajax-request
     * @param {Object} data - object of data
     * @param {Object} options - object of method options
     * @param {String} method - type of the request (get, post)
     * @returns {Promise} ajax request
     */
    Plugin.prototype.ajaxRequest = function (data, options = {}, method = 'post') {
        const defaultOptions = {
            processData: false,
            contentType: false,
            enctype: 'multipart/form-data',
            scrollingTopOnUpdate: Boolean(this.args.settings.scrolling_top_on_update),
            scrollingTopGap: Number(this.args.settings.scrolling_up_gap),
            scrollingTopSpeed: 500
        };

        options = $.extend({}, defaultOptions, options);

        // save extra parameters
        $.each(this.args, (key, value) => data.append(key, value));

        this.$element.addClass('is-loading');

        clearTimeout(this.getPreliminaryResultsCountTimeout);

        if (this.getPreliminaryResultsCountPromise && this.getPreliminaryResultsCountPromise.abort) {
            this.getPreliminaryResultsCountPromise.abort();
        }

        this.$document.trigger('ajaxRequest.wcsf', [this, data]);

        return $.ajax({
            url: this.args.ajaxUrl,
            method,
            data,
            processData: options.processData,
            contentType: options.contentType,
            enctype: options.enctype,
            cache: false,
            success: (response) => {
                // make redirect only
                if (response.redirect) {
                    this.options.documentNode.location = response.redirect;

                    return this;
                }

                this.$element.html(response.content).removeClass('is-loading');

                // scroll to top
                if (options.scrollingTopOnUpdate) {
                    let topPosition = this.$element.offset().top;

                    if (this.$window.scrollTop() > topPosition) {
                        if (options.scrollingTopGap) {
                            topPosition -= Number(options.scrollingTopGap);
                        }

                        this.$body.stop().animate({scrollTop: topPosition}, options.scrollingTopSpeed);
                    }
                }

                this.$document.trigger('ajaxCompleted.wcsf', [this, response]);

                return response;
            },
            error: (xhr, status, error) => {
                this.$element.removeClass('is-loading');

                window.console.error(xhr, status, error);

                return alert(`Unexpected error occurred: ${xhr.status}, ${xhr.statusText}`);
            }
        });
    };

    /**
     * Get current filter form data
     * @param {Object} data - extra data to pass
     * @returns {FormData} value
     */
    Plugin.prototype.getFormData = function (data = {}) {
        const defaultData = {referer: this.options.documentNode.location.href};

        data = $.extend({}, defaultData, data);

        const $form = this.$element.find('[data-component~="wcsf-form"]');
        const formData = new FormData($form.get(0));

        $.each(data, (key, value) => formData.append(key, typeof value !== 'string' ? JSON.stringify(value) : value));

        return formData;
    };

    /**
     * Submit step
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.submit = function (args = {}) {
        args.action = 'WCSFSubmit';

        return this.ajaxRequest(this.getFormData(args));
    };

    /**
     * Submit step and go next
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.submitAndGoNext = function (args = {}) {
        args.action = 'WCSFSubmitAndGoNext';

        return this.ajaxRequest(this.getFormData(args));
    };

    /**
     * Submit step and go to results
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.submitAndGoResults = function (args = {}) {
        args.action = 'WCSFSubmitAndGoResults';

        return this.ajaxRequest(this.getFormData(args));
    };

    /**
     * Submit step and get another one by id
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.submitAndGoTo = function (args = {}) {
        args.action = 'WCSFSubmitAndGoTo';

        return this.ajaxRequest(this.getFormData(args));
    };

    /**
     * Get step by id
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.getStep = function (args = {}) {
        args.action = 'WCSFGetStep';

        return this.ajaxRequest(this.getFormData(args));
    };

    /**
     * Reset form to the first step
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.reset = function (args = {}) {
        args.action = 'WCSFReset';

        return this.ajaxRequest(this.getFormData(args));
    };

    /**
     * Make a preliminary results query
     * @param {Object} $element - caller
     * @return {Plugin} self instance
     */
    Plugin.prototype.getPreliminaryResultsCount = function ($element) {
        const data = this.getFormData({action: 'WCSFGetPreliminaryResultsCount'});

        clearTimeout(this.getPreliminaryResultsCountTimeout);

        if (this.getPreliminaryResultsCountPromise && this.getPreliminaryResultsCountPromise.abort) {
            this.getPreliminaryResultsCountPromise.abort();
        }

        // save extra parameters
        $.each(this.args, (key, value) => data[key] = value);

        // pass cart to the server
        this.getPreliminaryResultsCountTimeout = setTimeout(
            () => {
                this.getPreliminaryResultsCountPromise = $.ajax({
                    url: this.args.ajaxUrl,
                    method: 'post',
                    data,
                    dataType: 'json',
                    processData: false,
                    contentType: false,
                    enctype: 'multipart/form-data',
                    cache: false,
                    success: (response) => this.$document
                        .trigger('getPreliminaryResultsCount.wcsf', [this, response, $element]),
                    error: (xhr, status, error) => {
                        if (status !== 'abort') {
                            window.console.error(xhr, status, error);
                        }
                    }
                });
            },
            this.options.preliminaryResultsCountRequestInterval
        );

        return this;
    };

    /**
     * Check HTML form validity
     * @param {Object} $form - jQuery object
     * @returns {Boolean} is valid
     */
    Plugin.prototype.checkFromValidity = function ($form = null) {
        let isValid = true;

        if (!$form || $form.length <= 0) {
            $form = this.$element.find('[data-component~="wcsf-form"]');
        }

        $form.each(function () {
            if (this.checkValidity !== 'undefined') {
                isValid = this.checkValidity();
            } else {
                $form.find(':input').each(function () {
                    const $input = $(this);

                    if ($input.attr('required') && $input.val() === '') {
                        isValid = false;
                    }
                });
            }
        });

        return isValid;
    };

    $.fn[pluginName] = function (options) {
        return this.each(function () {
            if (!$.data(this, 'wcsf')) {
                $.data(this, 'wcsf', new Plugin(this, options));
            }
        });
    };
});
